#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  5 10:46:47 2018

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import os

#::: allesfitter modules
from .. import config
from ..computer import update_params, calculate_model
from ..exoworlds_rdx.lightcurves import gp_decor
from ..exoworlds_rdx.lightcurves.index_transits import index_transits, index_eclipses
from ..exoworlds_rdx.rvs import estimate_jitter

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})






###############################################################################
#::: get a good initial guess for the baselines, errors & jitters
###############################################################################
def estimate_noise(datadir):
    global rv_err
    global data_minus_model
    
    
    #::: init
    config.init(datadir)
    
    #::: params
    params = update_params(config.BASEMENT.theta_0)
    
    #::: set up directory
    priordir = os.path.join(datadir,'priors')
    if not os.path.exists(priordir): os.makedirs(priordir)
    
      
    
    #::: set up phot summary file
    fname_summary = os.path.join(datadir,'priors','summary_phot.csv')
    with open( fname_summary, 'w+' ) as f:
        f.write('#name,gp_ln_sigma_median,gp_ln_sigma_ll,gp_ln_sigma_ul,gp_ln_rho_median,gp_ln_rho_ll,gp_ln_rho_ul,ln_yerr_median,ln_yerr_ll,ln_yerr_ul\n')
                  
    #::: run               
    for inst in config.BASEMENT.settings['inst_phot']:
        
        key = 'flux'
        print('\n###############################################################################')
        print(inst + ' ' +key)
        print('###############################################################################')
        outdir = os.path.join(datadir,'priors',inst)
        fname = inst+'_'+key
        
        time = config.BASEMENT.data[inst]['time']
        model = calculate_model(params, inst, key)
        data_minus_model = config.BASEMENT.data[inst][key] - model
        gp_decor(
                 time, data_minus_model, 
                 multiprocess=config.BASEMENT.settings['multiprocess'], multiprocess_cores=config.BASEMENT.settings['multiprocess_cores'],
                 outdir=outdir, fname=fname, fname_summary=fname_summary
                 )
                
    
        
    #::: set up rv summary file
    fname_summary = os.path.join(datadir,'priors','summary_rv.csv')
    with open( fname_summary, 'w+' ) as f:
        f.write('#name,ln_yerr_median,ln_yerr_ll,ln_yerr_ul\n')
                
    #::: run
    for inst in config.BASEMENT.settings['inst_rv']:
        
        key = 'rv'
        print('\n###############################################################################')
        print(inst + ' ' +key)
        print('###############################################################################')
        outdir = os.path.join(datadir,'priors',inst)
        fname = inst+'_'+key
                    
        time = config.BASEMENT.data[inst]['time']
        model = calculate_model(params, inst, key)
        data_minus_model = config.BASEMENT.data[inst][key] - model
        white_noise = config.BASEMENT.data[inst]['white_noise_'+key]
        
        estimate_jitter(
                        time, data_minus_model, white_noise,
                        multiprocess=config.BASEMENT.settings['multiprocess'], multiprocess_cores=config.BASEMENT.settings['multiprocess_cores'],
                        outdir=outdir, fname=fname, fname_summary=fname_summary
                        )
        




###############################################################################
#::: get a good initial guess for the baselines, errors & jitters
###############################################################################
def estimate_noise_out_of_transit(datadir, inst=None, **kwargs):
    
    #::: init
    config.init(datadir)
    
    #::: set up directory
    priordir = os.path.join(datadir,'priors')
    if not os.path.exists(priordir): os.makedirs(priordir)
    
    #::: set up phot summary file
    fname_summary = os.path.join(datadir,'priors','summary_phot.csv')
    with open( fname_summary, 'w+' ) as f:
        f.write('#name,gp_ln_sigma_median,gp_ln_sigma_ll,gp_ln_sigma_ul,gp_ln_rho_median,gp_ln_rho_ll,gp_ln_rho_ul,ln_yerr_median,ln_yerr_ll,ln_yerr_ul\n')
                  
    if inst is None:
        instruments = config.BASEMENT.settings['inst_phot']
    else:
        instruments = [inst]
        
    #::: run               
    for inst in instruments:
        
        key = 'flux'
        print('\n###############################################################################')
        print(inst + ' ' +key)
        print('###############################################################################')
        outdir = os.path.join(datadir,'priors',inst)
        fname = inst+'_'+key
        
        time = config.BASEMENT.fulldata[inst]['time']
        flux = config.BASEMENT.fulldata[inst][key]
        
        ind_in = []
        for companion in config.BASEMENT.settings['companions_phot']:
            epoch  = config.BASEMENT.params[companion+'_epoch']
            period = config.BASEMENT.params[companion+'_period']
            width  = config.BASEMENT.settings['fast_fit_width']
            if config.BASEMENT.settings['secondary_eclipse']:
                ind_ecl1, ind_ecl2, _ = index_eclipses(time,epoch,period,width,width) #TODO: currently this assumes width_occ == width_tra
                ind_in += list(ind_ecl1)
                ind_in += list(ind_ecl2)
            else:
                buf = list(index_transits(time,epoch,period,width)[0])
                ind_in += buf
        ind_in = np.sort(np.unique(ind_in))
    
        ind_out = []
        for i in np.arange(0, len(time), 1):
            if i not in ind_in:
                ind_out.append(i)
            
        
#        flux_err = config.BASEMENT.fulldata[inst]['err_scales_flux']
#        X = np.column_stack((time, flux, np.ones_like(flux)))
#        np.savetxt(inst+'_outoftransit.csv', X, delimiter=',')
                
        gp_decor(
                 time, flux, 
                 ind_in = ind_in, ind_out=ind_out,
                 multiprocess=config.BASEMENT.settings['multiprocess'], multiprocess_cores=config.BASEMENT.settings['multiprocess_cores'],
                 outdir=outdir, fname=fname, fname_summary=fname_summary, 
                 **kwargs
                 )
    
        
        


###############################################################################
#::: old
###############################################################################
        #guess the total yerr = np.sqrt( rv_err**2 + jitter**2  )
#        yerr = np.nanstd(data_minus_model) * np.ones_like(data_minus_model)
        
        #re-calcualte an initital guess for log_jitter 
#        initial = np.nanmedian( np.log( np.sqrt( yerr**2 - rv_err**2 ) ) )
#        print(initial)
        
#        def neg_log_like(theta):
#            log_jitter = theta
#            jitter = np.exp(log_jitter)
#            yerr = np.sqrt( rv_err**2 + jitter**2 )
#            inv_sigma2_w = 1./yerr**2
#            return +0.5*(np.nansum((data_minus_model)**2 * inv_sigma2_w - np.log(inv_sigma2_w)))
        
        
        #::: Diff. Evol.
#        bounds = [(-23,0)]
#        soln_DE = differential_evolution(neg_log_like, bounds)      

#        print(inst, key)
##        print('\tguess:', initial, 'lnlike:', neg_log_like(initial))
#        print('\tDE:', soln_DE.x[0], 'lnlike:', neg_log_like(soln_DE.x[0]))
        

    
    
###############################################################################
#::: old 2
###############################################################################
#def get_initial_guess_gp(datadir):
#    import celerite
#    from celerite import terms
#    from scipy.optimize import minimize
#    from .computer import update_params
#    
#    config.init(datadir)
#    base = config.BASEMENT
#    params = update_params(base.theta_0)
#    
#    for inst in base.settings['inst_phot']:
#        key = 'flux'
#        model = calculate_model(params, inst, key, xx=None)
#        x = base.data[inst]['time']
#        y = base.data[inst][key] - model
##        yerr_weights = config.BASEMENT.data[inst]['err_scales_'+key]
#        yerr = np.nanstd(y) #overwrite yerr; works better for removing smooth global trends
#        
#        kernel = terms.Matern32Term(log_sigma=1., log_rho=1.)
#        gp = celerite.GP(kernel, mean=np.nanmean(y)) 
#        gp.compute(x, yerr=yerr) #constrain on x/y/yerr
#         
#        def neg_log_like(gp_params, y, gp):
#            gp.set_parameter_vector(gp_params)
#            return -gp.log_likelihood(y)
#        
#        def grad_neg_log_like(gp_params, y, gp):
#            gp.set_parameter_vector(gp_params)
#            return -gp.grad_log_likelihood(y)[1]
#        
#        initial_params = gp.get_parameter_vector()
#        bounds = gp.get_parameter_bounds()
#        soln = minimize(neg_log_like, initial_params, jac=grad_neg_log_like,
#                        method="L-BFGS-B", bounds=bounds, args=(y, gp))
##        gp.set_parameter_vector(soln.x)
#        
#        inv_sigma2_w = calculate_inv_sigma2_w(params, inst, key)
#    
#        print('baseline_gp1_'+key+'_'+inst + ':', soln.x[0])
#        print('baseline_gp2_'+key+'_'+inst + ':', soln.x[1])
#        print('inv_sigma2_'+key+'_'+inst + ':', np.nanmean(inv_sigma2_w))



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  5 01:03:21 2018

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import os
import emcee
import multiprocessing
multiprocessing.set_start_method('fork', force=True)
#solves python>=3.8 issues, see https://stackoverflow.com/questions/60518386/error-with-module-multiprocessing-under-python3-8
from multiprocessing import Pool
from contextlib import closing
from time import time as timer

#::: warnings
import warnings
warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning) 
warnings.filterwarnings('ignore', category=np.RankWarning) 

#::: allesfitter modules
from . import config
from .computer import update_params, calculate_lnlike_total
from .general_output import logprint
from .mcmc_output import print_autocorr




###############################################################################
#::: MCMC log likelihood
###############################################################################
def mcmc_lnlike(theta):
    
    params = update_params(theta)
    lnlike = calculate_lnlike_total(params)
    
#    lnlike = 0
#    
#    for inst in config.BASEMENT.settings['inst_phot']:
#        lnlike += calculate_lnlike(params, inst, 'flux')
#    
#    for inst in config.BASEMENT.settings['inst_rv']:
#        lnlike += calculate_lnlike(params, inst, 'rv')
#        
#    if np.isnan(lnlike) or np.isinf(lnlike):
#        lnlike = -np.inf

    return lnlike

    
        
###############################################################################
#::: MCMC log prior
###############################################################################
def mcmc_lnprior(theta):
    '''
    bounds has to be list of len(theta), containing tuples of form
    ('none'), ('uniform', lower bound, upper bound), or ('normal', mean, std)
    '''
    lnp = 0.        
    
    for th, b in zip(theta, config.BASEMENT.bounds):
        if b[0] == 'uniform':
            if not (b[1] <= th <= b[2]): 
                return -np.inf
        elif b[0] == 'normal':
            lnp += np.log( 1./(np.sqrt(2*np.pi) * b[2]) * np.exp( - (th - b[1])**2 / (2.*b[2]**2) ) )
        elif b[0] == 'trunc_normal':
            if not (b[1] <= th <= b[2]): 
                return -np.inf
            lnp += np.log( 1./(np.sqrt(2*np.pi) * b[4]) * np.exp( - (th - b[3])**2 / (2.*b[4]**2) ) )
        else:
            raise ValueError('Bounds have to be "uniform" or "normal". Input from "params.csv" was "'+b[0]+'".')
    return lnp



###############################################################################
#::: MCMC log probability
###############################################################################    
def mcmc_lnprob(theta):
    '''
    has to be top-level for  for multiprocessing pickle
    '''
    lp = mcmc_lnprior(theta)
    if not np.isfinite(lp):
        return -np.inf
    else:
#        try:
        ln = mcmc_lnlike(theta)
        return lp + ln
#        except:
#            return -np.inf
        


###########################################################################
#::: MCMC fit
###########################################################################
def mcmc_fit(datadir):
    
    #::: init
    config.init(datadir)
    
    
    continue_old_run = False
    if os.path.exists(os.path.join(config.BASEMENT.outdir,'mcmc_save.h5')):
        overwrite = str(input(os.path.join(config.BASEMENT.outdir,'mcmc_save.h5')+\
                              ' already exists.\n'+\
                              'What do you want to do?\n'+\
                              '1 : overwrite the save file\n'+\
                              '2 : append to the save file\n'+\
                              '3 : abort\n'))
        if (overwrite == '1'):
            continue_old_run = False
        elif (overwrite == '2'):
            continue_old_run = True
        else:
            raise ValueError('User aborted operation.')
            
    
    #::: continue on the backend / reset the backend
    if os.path.exists(os.path.join(config.BASEMENT.outdir,'mcmc_save.h5')) and not continue_old_run:
        #backend.reset(config.BASEMENT.settings['mcmc_nwalkers'], config.BASEMENT.ndim)
        os.remove(os.path.join(config.BASEMENT.outdir,'mcmc_save.h5'))
        
        
    #::: set up a fresh backend
    backend = emcee.backends.HDFBackend(os.path.join(config.BASEMENT.outdir,'mcmc_save.h5'))
    
    
    #::: helper fct
    def run_mcmc(sampler):
        
        #::: set initial walker positions
        if continue_old_run:
            p0 = backend.get_chain()[-1,:,:]
            already_completed_steps = backend.get_chain().shape[0] * config.BASEMENT.settings['mcmc_thin_by']
        else:
            p0 = config.BASEMENT.theta_0 + config.BASEMENT.init_err * np.random.randn(config.BASEMENT.settings['mcmc_nwalkers'], config.BASEMENT.ndim)
            already_completed_steps = 0
        
        #::: make sure the inital positions are within the limits
        for i, b in enumerate(config.BASEMENT.bounds):
            if b[0] == 'uniform':
                p0[:,i] = np.clip(p0[:,i], b[1], b[2]) 
        
        #::: if pre-runs are wished for, and if we are not continuing an existing run
        if continue_old_run==False:
            for i in range(config.BASEMENT.settings['mcmc_pre_run_loops']):
                logprint("\nRunning pre-run loop",i+1,'/',config.BASEMENT.settings['mcmc_pre_run_loops'])
                
                #::: run the sampler        
                sampler.run_mcmc(p0,
                                 config.BASEMENT.settings['mcmc_pre_run_steps'],
                                 progress=config.BASEMENT.settings['print_progress'])
                
                #::: get maximum likelhood solution
                log_prob = sampler.get_log_prob(flat=True)
                posterior_samples = sampler.get_chain(flat=True)
                ind_max = np.argmax(log_prob)
                p0 = posterior_samples[ind_max,:] + config.BASEMENT.init_err * np.random.randn(config.BASEMENT.settings['mcmc_nwalkers'], config.BASEMENT.ndim)
    
                #::: reset sampler and backend
                #backend.reset(config.BASEMENT.settings['mcmc_nwalkers'], config.BASEMENT.ndim)
                os.remove(os.path.join(config.BASEMENT.outdir,'mcmc_save.h5'))
                sampler.reset()
        
        #::: run the sampler        
        logprint("\nRunning full MCMC")
        sampler.run_mcmc(p0,
                         int((config.BASEMENT.settings['mcmc_total_steps'] - already_completed_steps)/config.BASEMENT.settings['mcmc_thin_by']), 
                         thin_by=int(config.BASEMENT.settings['mcmc_thin_by']), 
                         progress=config.BASEMENT.settings['print_progress'])
        
        return sampler


    #::: Run
    logprint("\nRunning MCMC...")
    logprint('--------------------------')
    t0 = timer()
    if config.BASEMENT.settings['multiprocess']:
        with closing(Pool(processes=(config.BASEMENT.settings['multiprocess_cores']))) as pool: #multiprocessing
#        with closing(Pool(cpu_count()-1)) as pool: #pathos
            logprint('\nRunning on', config.BASEMENT.settings['multiprocess_cores'], 'CPUs.')
            sampler = emcee.EnsembleSampler(config.BASEMENT.settings['mcmc_nwalkers'], 
                                            config.BASEMENT.ndim, 
                                            mcmc_lnprob,
                                            moves=config.BASEMENT.settings['mcmc_moves'],
                                            pool=pool, 
                                            backend=backend)
            sampler = run_mcmc(sampler)
        t1 = timer()
        timemcmc = (t1-t0)
        logprint("\nTime taken to run 'emcee' on", config.BASEMENT.settings['multiprocess_cores'], "cores is {:.2f} hours".format(timemcmc/60./60.))
        
    else:
        sampler = emcee.EnsembleSampler(config.BASEMENT.settings['mcmc_nwalkers'],
                                        config.BASEMENT.ndim,
                                        mcmc_lnprob,
                                        moves=config.BASEMENT.settings['mcmc_moves'],
                                        backend=backend)
        sampler = run_mcmc(sampler)
        t1 = timer()
        timemcmc = (t1-t0)
        logprint("\nTime taken to run 'emcee' on a single core is {:.2f} hours".format(timemcmc/60./60.))
    

    #::: Check performance and convergence
    logprint('\nAcceptance fractions:')
    logprint('--------------------------')
    logprint(sampler.acceptance_fraction)
    
    print_autocorr(sampler)
    
    
    #::: return a German saying
    try:
        with open(os.path.join(os.path.dirname(__file__), 'utils', 'quotes2.txt')) as dataset:
            return(np.random.choice([l for l in dataset]))
    except:
        return('42')
    
    


# -*- coding: utf-8 -*-
"""
Created on Wed Apr 27 12:55:39 2016

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np

#::: lightcurves modules
from .utils import mask_ranges




def get_first_epoch(time, epoch, period, width=0):    
    '''
    width : float
        set >0 to include transit egress to mark the first transit
    place the first_epoch at the start of the data to avoid luser mistakes
    '''
    time = np.sort(np.atleast_1d(time))
    start = np.nanmin( time )
    first_epoch = 1.*epoch + width/2. #add width/2 to catch egress
    if start<=first_epoch: first_epoch -= np.floor((first_epoch-start)/period) * period
    else: first_epoch += np.ceil((start-first_epoch)/period) * period
    return first_epoch - width/2.  #subtract width/2 to get midpoint again
    
    

def index_transits(time, epoch, period, width):
    """
    Returns:
    --------
    ind_tr : array
        indices of points in transit
    ind_out : array
        indices of points out of transit
    """
    time = np.sort(time)
    epoch = get_first_epoch(time, epoch, period, width=width)
    try:
        N = int( 1. * ( time[-1] - epoch ) / period ) + 2
        tmid = np.array( [ epoch + i * period for i in range(N) ] )
        _, ind_tr, mask_tr = mask_ranges( time, tmid - width/2., tmid + width/2. )
        ind_out = np.arange( len(time) )[ ~mask_tr ]
    except:
        ind_tr, ind_out = slice(0), slice(None)

    return ind_tr, ind_out 
    
    

#::: for binaries, mark the primary and secondary eclipse
#TODO: implement non-circular orbits
def index_eclipses(time, epoch, period, width_1, width_2):
    """
    Returns:
    --------
    ind_ecl1 : array
        indices of points in primary eclipse
    ind_ecl2 : array
        indices of points in secondary eclipse
    ind_out : array
        outside of any eclipse
    
    ! this assumes circular orbits !
    """
    time = np.sort(time)
    epoch = get_first_epoch(time, epoch, period, width=width_1)
    N = int( 1. * ( time[-1] - epoch ) / period ) + 1
        
    tmid_ecl1 = np.array( [ epoch +             i * period  for i in range(N) ] )
    tmid_ecl2 = np.array( [ epoch - period/2. + i * period  for i in range(N+1) ] )
    
    _, ind_ecl1, mask_ecl1 = mask_ranges( time, tmid_ecl1 - width_1/2., tmid_ecl1 + width_1/2. )           
    _, ind_ecl2, mask_ecl2 = mask_ranges( time, tmid_ecl2 - width_2/2., tmid_ecl2 + width_2/2. )
        
    ind_out = np.arange( len(time) )[ ~(mask_ecl1 | mask_ecl2) ]

    return ind_ecl1, ind_ecl2, ind_out
    


def index_eclipses_smart(time, epoch, period, rr, rsuma, cosi, f_s, f_c, extra_factor=1.):
    #--------------------------------------------------------------------------
    #::: inputs
    #--------------------------------------------------------------------------
    R_star_over_a = rsuma / (1. + rr)
    ecc = f_s**2 + f_c**2
    esinw = f_s * np.sqrt(ecc)
    ecosw = f_c * np.sqrt(ecc)
    
    
    #--------------------------------------------------------------------------
    #::: widths and impact parameters 
    #--------------------------------------------------------------------------
    eccentricity_correction_b_1 = ( (1. - ecc**2) / ( 1. + esinw ) )
    b_1 = (1./R_star_over_a) * cosi * eccentricity_correction_b_1

    eccentricity_correction_width_1 = ( np.sqrt(1. - ecc**2) / ( 1. + esinw ) )
    width_1 = period/np.pi  \
            * np.arcsin( R_star_over_a \
                         * np.sqrt( (1. + rr)**2 - b_1**2  )\
                         / np.sin(np.arccos(cosi)) ) \
            * eccentricity_correction_width_1
            
    eccentricity_correction_b_2 = ( (1. - ecc**2) / ( 1. - esinw ) )
    b_2 = (1./R_star_over_a) * cosi * eccentricity_correction_b_2
    
    width_2 = width_1 * (1. + esinw) / (1. - esinw)
    print(width_1, width_2)
    print(b_1, b_2)
    
    
    #--------------------------------------------------------------------------
    #::: timing
    #--------------------------------------------------------------------------
    time = np.sort(time)
    epoch = get_first_epoch(time, epoch, period, width=2*width_1)
    epoch_occ = epoch + period/2. * (1. + 4./np.pi * ecosw)
    
    N = int( 1. * ( time[-1] - epoch ) / period ) + 1
        
    tmid_ecl1 = np.array( [ epoch + i * period for i in range(N) ] )
    tmid_ecl2 = np.array( [ epoch_occ + (i-1) * period for i in range(N+1) ] ) #(o-1) in case an occ happens before the first epoch
    
    _, ind_ecl1, mask_ecl1 = mask_ranges( time, tmid_ecl1 - width_1/2.*extra_factor, tmid_ecl1 + width_1/2.*extra_factor )           
    _, ind_ecl2, mask_ecl2 = mask_ranges( time, tmid_ecl2 - width_2/2.*extra_factor, tmid_ecl2 + width_2/2.*extra_factor )
        
    ind_out = np.arange( len(time) )[ ~(mask_ecl1 | mask_ecl2) ]
    
    
    #--------------------------------------------------------------------------
    #::: return
    #--------------------------------------------------------------------------
    return ind_ecl1, ind_ecl2, ind_out



def get_tmid_transits(time, epoch, period, width):
    '''
    get a list of only the transit midpoints that are actually covered by the data
    '''
    time = np.sort(time)
    epoch = get_first_epoch(time, epoch, period, width=width)
    N = int( 1. * ( time[-1] - epoch ) / period ) + 1
    tmid = np.array( [ epoch + i * period for i in range(N) ] )
    
    return tmid


    
def get_tmid_observed_transits(time, epoch, period, width):
    '''
    get a list of only the transit midpoints that are actually covered by the data
    '''
    time = np.sort(time)
    epoch = get_first_epoch(time, epoch, period, width=width)
    N = int( 1. * ( time[-1] - epoch ) / period ) + 1
    tmid = np.array( [ epoch + i * period for i in range(N) ] )
    
    tmid_observed_transits = []
    
    for i,t in enumerate(tmid): 
        mask = ((time >= (t - width/2.)) & (time <= (t + width/2.)))
        if any(mask): 
            tmid_observed_transits.append( tmid[i] )
    
    return tmid_observed_transits




#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  5 14:18:20 2018

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

from .index_transits import index_transits, index_eclipses, get_first_epoch
from .lightcurve_tools import phase_fold, rebin_err
from .expand_flags import expand_flags
from .gp_decor import gp_decor



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 26 15:36:21 2019

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import os

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




def print_priors(datadir, typ='uniform', scaling=5):

    if os.path.exists( os.path.join(datadir,'results/ns_table.csv') ):
        f = os.path.join(datadir,'results/ns_table.csv')
    else:
        f = os.path.join(datadir,'results/mcmc_table.csv')
        
    results = np.genfromtxt( f, delimiter=',', dtype=('U100',float,float,float,'U100','U100'), encoding='utf-8', names=True )

    if typ=='uniform':
        for i in range(len(results['name'])):
            if np.isnan(results['lower_error'][i]):
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',0,,' + results['label'][i] + ',' + results['unit'][i])
            elif ('_period' in results['name'][i]) or ('_epoch' in results['name'][i]) or ('_slope' in results['name'][i]) or ('_offset' in results['name'][i]):
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',1,uniform ' + str(results['median'][i] - scaling*results['lower_error'][i]) + ' ' + str(results['median'][i] + scaling*results['upper_error'][i])  + ',' + results['label'][i] + ',' + results['unit'][i])
            elif ('_f_c' in results['name'][i]) or ('_f_s' in results['name'][i]):
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',1,uniform ' + str( np.max( [-1, results['median'][i] - scaling*results['lower_error'][i]] ) ) + ' ' + str( np.min( [1, results['median'][i] + scaling*results['upper_error'][i]] ) )  + ',' + results['label'][i] + ',' + results['unit'][i])
            elif ('baseline_gp' in results['name'][i]):
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',1,uniform ' + str( np.max( [-15, results['median'][i] - scaling*results['lower_error'][i]] ) ) + ' ' + str( np.min( [15, results['median'][i] + scaling*results['upper_error'][i]] ) )  + ',' + results['label'][i] + ',' + results['unit'][i])
            elif ('ln_err' in results['name'][i]) or ('ln_jitter' in results['name'][i]):
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',1,uniform ' + str( np.max( [-15, results['median'][i] - scaling*results['lower_error'][i]] ) ) + ' ' + str( np.min( [0, results['median'][i] + scaling*results['upper_error'][i]] ) )  + ',' + results['label'][i] + ',' + results['unit'][i])
            else:
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',1,uniform ' + str( np.max( [0, results['median'][i] - scaling*results['lower_error'][i]] ) ) + ' ' + str( np.min( [1, results['median'][i] + scaling*results['upper_error'][i]] ) )  + ',' + results['label'][i] + ',' + results['unit'][i])

    if typ=='trunc_normal':
        for i in range(len(results['name'])):
            if np.isnan(results['lower_error'][i]):
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',0,,' + results['label'][i] + ',' + results['unit'][i])
            elif ('_period' in results['name'][i]) or ('_epoch' in results['name'][i]) or ('_slope' in results['name'][i]) or ('_offset' in results['name'][i]):
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',1,trunc_normal 0 1e15 ' + str(results['median'][i]) + ' ' + scaling*str(np.max([results['lower_error'][i],results['upper_error'][i]])) + ',' + results['label'][i] + ',' + results['unit'][i])
            elif ('_f_c' in results['name'][i]) or ('_f_s' in results['name'][i]):
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',1,trunc_normal -1 1 ' + str(results['median'][i]) + ' ' + scaling*str(np.max([results['lower_error'][i],results['upper_error'][i]])) + ',' + results['label'][i] + ',' + results['unit'][i])
            elif ('baseline_gp' in results['name'][i]):
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',1,trunc_normal -15 15 ' + str(results['median'][i]) + ' ' + scaling*str(np.max([results['lower_error'][i],results['upper_error'][i]])) + ',' + results['label'][i] + ',' + results['unit'][i])
            elif ('ln_err' in results['name'][i]) or ('ln_jitter' in results['name'][i]):
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',1,trunc_normal -15 0 ' + str(results['median'][i]) + ' ' + scaling*str(np.max([results['lower_error'][i],results['upper_error'][i]])) + ',' + results['label'][i] + ',' + results['unit'][i])
            else:
                print(str(results['name'][i]) + ',' + str(results['median'][i]) + ',1,trunc_normal 0 1 ' + str(results['median'][i]) + ' ' + scaling*str(np.max([results['lower_error'][i],results['upper_error'][i]])) + ',' + results['label'][i] + ',' + results['unit'][i])




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Jul 10 18:27:01 2020

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
from matplotlib.colors import ColorConverter
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
from matplotlib.gridspec import GridSpec
from mpl_toolkits.axes_grid1 import make_axes_locatable
import colorsys
import warnings
import pandas as pd
from tqdm import tqdm
from glob import glob
from pprint import pprint
try: 
    from brokenaxes import brokenaxes
except: 
    pass #don't throw an error for now, otherwise people will panic
import pathlib
from astropy.time import Time
from datetime import date

#::: local imports
from allesfitter.inout import read_csv
from allesfitter.time_series import sigma_clip, slide_clip

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in", "ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth':1})



'''
This script offers several convenience functions for plotting diverse time series data, 
and particularly lightcurves. These are, in order of generality:
    - fullplot
    - brokenplot
    - chunkplot
    - monthplot
    - tessplot
'''
    


def fullplot(time, y, yerr=None, clip=False, ax=None, **kwargs):
    '''
    Parameters
    ----------
    time : array of float
        Time stamps, can be in any format of days.
    y : array of float
        e.g. y or RV array (usually as normalized y or RV in km/s).
    yerr : array of float
        e.g. y or RV error array (usually as normalized y or RV in km/s).
    ax : matplotlib axis, optional
        An existing figure axis. The default is None.
    clip : bool, optional
        Automatically clip the input data with sigma_clip(low=4, high=4)
        and slide_clip(window_length=1, low=4, high=4). The default is True.

    Returns
    -------
    ax : matplotlib axis
        A single figure axis.
    '''
    #::: initialize
    time, y, yerr, ax, kwargs, mask_lower, mask_upper = _initialize(time, y, yerr, ax, clip, kwargs)
    
    #::: plot
    _plot1(ax, slice(None), time, y, yerr, mask_upper, mask_lower, clip, **kwargs)
        
    #::: make the axes pretty
    _set_axes(ax, time, y)
    
    #::: return
    return ax



def fullplot_csv(fname, clip=False, ax=None):
    '''
    Wrapper around fullplot to plot straight from a csv file.
    See fullplot() for details.
    '''
    time, y, yerr = read_csv(fname)[0:3]
    return fullplot(time, y, yerr, clip=clip, ax=ax)



def brokenplot(time, y, yerr=None, dt=10, clip=False, bax=None, **kwargs):
    '''
    Parameters
    ----------
    time : array of float
        e.g. time array (usually in days).
    y : array of float
        e.g. y or RV array (usually as normalized y or RV in km/s).
    yerr : array of float
        e.g. y or RV error array (usually as normalized y or RV in km/s).
    dt : float, optional
        The gap size after which axes will be broken. The default is 10 (usually in days).
    bax : brokenaxes instance, optional
        An existing brokenaxis axis. Note that this cannot be a normal matplotlib axis. 
        The default is None.
    time_format : str
        The format of your time array. 
        Must be either 'BJD_TDB' or 'TJD' (TESS Julian Date). 
        The default is 'BJD_TDB'.
    clip : bool, optional
        Automatically clip the input data with sigma_clip(low=4, high=4)
        and slide_clip(window_length=1, low=4, high=4). The default is True.

    Returns
    -------
    bax : brokenaxes instance
        Just like an pyplot.Axes instance.
    '''
    #::: check if brokenaxes is installed
    if 'brokenaxes' not in sys.modules:
        raise ImportError('You must install the brokenaxes package before using allesfitter.brokenplot()\n(pip install brokenaxes)')
    
    #::: initialize (but skip the figure/axes creation here)
    time, y, yerr, _, kwargs, mask_lower, mask_upper = _initialize(time, y, yerr, 1, clip, kwargs)
    
    #::: special setup for brokenaxes
    if bax is None:
        plt.figure(figsize=(12,3))    
        ind0 = [0] + list(np.where(np.diff(time)>dt)[0]+1) #start indices of data chunks
        ind1 = list(np.where(np.diff(time)>dt)[0]) + [len(time)-1] #end indices of data chunks
        xlims = [ (time[i]-(time[j]-time[i])/100.,time[j]+(time[j]-time[i])/100.) for i,j in zip(ind0,ind1) ]
        bax = brokenaxes(xlims=xlims, wspace=0.05)
        
    #::: plot    
    _plot1(bax, slice(None), time, y, yerr, mask_upper, mask_lower, clip, is_bax=True, **kwargs)

    #::: make the axes pretty (special one necessary here)
    _set_brokenaxes(bax, time, y)  
    
    #::: return
    return bax



def brokenplot_csv(fname, dt=10, clip=False, bax=None, **kwargs):
    '''
    Wrapper around brokenplot to plot straight from a csv file.
    See brokenplot() for details.
    '''
    time, y, yerr = read_csv(fname)[0:3]
    return brokenplot(time, y, yerr, dt=dt, clip=clip, bax=bax, **kwargs)



def chunkplot(time, y, yerr=None, chunk_size=30., clip=False, sharey=False, axes=None, **kwargs):
    '''
    Creates a new line for every chunk of days determined by chunk_size, with
    the count starting at the first time stamp. The time format can be arbitrary here.
    
    Parameters
    ----------
    time : array of float
        Time stamps, can be in any format of days.
    y : array of float
        y.
    y_err : array of float, optional
        y error bars. The default is None.
    chunk_size : float
        The length of each chunk of days to be plotted in a single axis. 
        The default is 30., i.e. roughly a month.
    clip : bool, optional
        Automatically clip the input data with sigma_clip(low=4, high=4)
        and slide_clip(window_length=1, low=4, high=4). The default is True.
    sharey : bool
        Share the y-axis between all rows. The default is True.
    kwargs : ...
        Any keyword arguments you wish to pass to the plot function (e.g., color='red')

    Returns
    -------
    None.
    '''
    #::: initialize (but skip the figure/axes creation here)
    time, y, yerr, _, kwargs, mask_lower, mask_upper = _initialize(time, y, yerr, 1, clip, kwargs)
    
    #::: get chunks of chunk_size days (in the same units of time)
    start_of_chunk = [ np.nanmin(time)+i*chunk_size for i in range( 1+int(1.*np.ptp(time)/chunk_size) ) ]
    end_of_chunk = [ x + chunk_size - 1e-12 for x in start_of_chunk ]
    start_of_chunk_str = [ 'Chunk '+str(i+1) for i in range(len(start_of_chunk)) ]
    
    #::: if this is a brand new figure, match the init_time to TESS orbits and sectors and then set up all the axes
    if axes is None:
            
        #::: select which months are covered by the data
        chunks, xlims0, xlims1  = [], [], [],
        for i in range(len(start_of_chunk)):
            t0 = start_of_chunk[i]
            t1 = end_of_chunk[i]
            ind = np.where( (time > t0) & (time <= t1) )[0]
            if len(ind)>0:
                chunks.append( start_of_chunk_str[i] )
                xlims0.append( start_of_chunk[i] )
                xlims1.append( end_of_chunk[i] )
    
        #::: set up the figure
        N = len(chunks)
        fig, axes = plt.subplots(N, figsize=(12,3*N), tight_layout=True, sharey=sharey)
            
        #::: loop over all months/axes to set up each axis
        for i, c in enumerate(chunks):
            ax1 = np.atleast_1d(axes)[i]
            
            #::: add text
            ax1.text(0.98, 0.95, c, ha='right', va='top', transform=ax1.transAxes, zorder=15)
            
            #::: set the axis xlim and labels properly
            ax1.set(xlim=[xlims0[i], xlims1[i]])    
            _set_axes(ax1, time, y, labels={'show_x': (i==N-1)})    
        
    #::: iterate over all existing, set-up axes and plot
    for i, ax1 in enumerate(np.atleast_1d(axes)):
        
        #::: check if there is data to be plotted on this axis
        ind = np.where( (time > ax1.get_xlim()[0]) & (time <= ax1.get_xlim()[1]) )[0]
        
        #::: only proceed if there is actually data in this range
        if len(ind)>0:
            
            #::: plot (check if it should be a scatter or errorbar plot)
            _plot1(ax1, ind, time, y, yerr, mask_upper, mask_lower, clip, **kwargs)
            
    #::: return
    return axes



def chunkplot_csv(fname, **kwargs):
    '''
    Wrapper around tessplot to plot straight from a csv file.
    See tessplot() for details.
    '''
    time, y, yerr = read_csv(fname)[0:3]
    return chunkplot(time, y, yerr, **kwargs)  



def monthplot(time, y, yerr=None, clip=False, sharey=False, axes=None, **kwargs):
    '''
    Creates a new line for every calendar month.
    The time format must be BJD_TDB here.
    
    Parameters
    ----------
    time : array of float
        Time stamps, must be in BJD_TDB (do not use arbitrary days nor TJD).
    y : array of float
        Relative flux, RV, or residuals.
    yerr : array of float, optional
        Error bars of the y values. The default is None.
    clip : bool, optional
        Automatically clip the input data with sigma_clip(low=4, high=4)
        and slide_clip(window_length=1, low=4, high=4). The default is True.
    sharey : bool
        Share the y-axis between all rows. The default is True.
    kwargs : ...
        Any keyword arguments you wish to pass to the plot function (e.g., color='red')

    Returns
    -------
    None.
    '''
    #::: sanity check the data
    jd_moon_landing = 2440422.5 #(fake)
    if np.nanmin(time) < jd_moon_landing:
        raise ValueError('The time array does not seem to be in units of BJD_TDB.')
        
    #::: initialize (but skip the figure/axes creation here)
    time, y, yerr, _, kwargs, mask_lower, mask_upper = _initialize(time, y, yerr, 1, clip, kwargs)
    
    #::: get the months since the moon landing (fake)
    start_of_month = [str(year)+'-'+str(month).zfill(2)+'-01T00:00:00' 
                      for year in range(1969,date.today().year+1) for month in range(1,13)]
    start_of_month = Time(start_of_month, format='isot', scale='utc')
    start_of_month_str = [t.strftime('%Y %b') for t in start_of_month]
    
    #::: if this is a brand new figure, match the init_time to TESS orbits and sectors and then set up all the axes
    if axes is None:
            
        #::: select which months are covered by the data
        months, xlims0, xlims1  = [], [], [],
        for i in range(len(start_of_month)-1):
            t0 = start_of_month.jd[i]
            t1 = start_of_month.jd[i+1]-1e-12
            ind = np.where( (time > t0) & (time <= t1) )[0]
            if len(ind)>0:
                months.append( start_of_month_str[i] )
                xlims0.append( start_of_month.jd[i] )
                xlims1.append( start_of_month.jd[i+1]-1e-12 )
    
        #::: set up the figure
        N = len(months)
        fig, axes = plt.subplots(N, figsize=(12,3*N), tight_layout=True, sharey=sharey)
            
        #::: loop over all months/axes to set up each axis
        for i, m in enumerate(months):
            ax1 = np.atleast_1d(axes)[i]
            
            #::: add text
            ax1.text(0.98, 0.95, m, ha='right', va='top', transform=ax1.transAxes, zorder=15)
            
            #::: set the axis xlim and labels properly
            ax1.set(xlim=[xlims0[i], xlims1[i]])    
            _set_axes(ax1, time, y, labels={'show_x': (i==N-1)})    
        
    #::: iterate over all existing, set-up axes and plot
    for i, ax1 in enumerate(np.atleast_1d(axes)):
        
        #::: check if there is data to be plotted on this axis
        ind = np.where( (time > ax1.get_xlim()[0]) & (time <= ax1.get_xlim()[1]) )[0]
        
        #::: only proceed if there is actually data in this range
        if len(ind)>0:
            
            #::: plot (check if it should be a scatter or errorbar plot)
            _plot1(ax1, ind, time, y, yerr, mask_upper, mask_lower, clip, **kwargs)
    
    #::: return
    return axes
    


def monthplot_csv(fname, **kwargs):
    '''
    Wrapper around tessplot to plot straight from a csv file.
    See tessplot() for details.
    '''
    time, y, yerr = read_csv(fname)[0:3]
    return monthplot(time, y, yerr, **kwargs)    
    


def tessplot(time, y, yerr=None, clip=False, sharey=False, axes=None, **kwargs):
    '''
    A collage plot for long TESS lightcurves. Creates a new axes row for every TESS Sector.
    The time input must be in BJD_TDB. The first call to this function must fully initialize it,
    i.e. it must cover all time stamps used to select the right sectors and set the axes properly.
    If your first call of tessplot only uses a fraction of the full data and you plan to add more later,
    you must use init_time to pass the correct, full time stamps to correctly initialize the figure and all axes.
    
    Parameters
    ----------
    time : array of float
        Time stamps, must be in BJD_TDB (do not use arbitrary days nor TJD).
    y : array of float
        Relative flux, RV, or residuals.
    yerr : array of float, optional
        Error bars of the y values. The default is None.
    clip : bool, optional
        Automatically clip the input data with sigma_clip(low=4, high=4)
        and slide_clip(window_length=1, low=4, high=4). The default is False.
    sharey : bool
        Share the y-axis between all rows. The default is False.
    axes : matplotlib axes, optional
        Uses existing figure axes instead of making a new figure. The default is None.
    kwargs : ...
        Any keyword arguments you wish to pass to the plot function (e.g., color='r')

    Returns
    -------
    ax : matplotlib axes
        The figure axes.
    '''
    #::: sanity check the data
    jd_moon_landing = 2440422.5 #(fake)
    if np.nanmin(time) < jd_moon_landing:
        raise ValueError('The time array does not seem to be in units of BJD_TDB.')
    
    #::: initialize (but skip the figure/axes creation here)
    time, y, yerr, _, kwargs, mask_lower, mask_upper = _initialize(time, y, yerr, 1, clip, kwargs)
    
    #::: load the TESS orbit times
    here = pathlib.Path(__file__).parent.absolute()
    df = pd.read_csv( os.path.join(here,'_static','_tess','tess_orbit_times.csv'), comment="#" )
    
    #::: if this is a brand new figure, match the init_time to TESS orbits and sectors and then set up all the axes
    if axes is None:
            
        #::: count which sectors are covered by init_time
        sectors, xlims0, xlims1 = [], [], []
        for s in range(1,max(df['Sector'])):
            line = 2*(s-1)
            t0 = Time(df['Start of Orbit'].loc[line]).jd
            t1 = Time(df['End of Orbit'].loc[line+1]).jd
            ind = np.where( (time > t0) & (time <= t1) )[0]
            if len(ind)>0:
                sectors.append(s)
                xlims0.append(t0-0.5)
                xlims1.append(t1+0.5)
    
        #::: set up the figure
        N = len(sectors)
        fig, axes = plt.subplots(N, figsize=(12,3*N), tight_layout=True, sharey=sharey)
            
        #::: loop over all sectors/axes to set up each axis
        for i, s in enumerate(sectors):
            line = 2*(s-1)
            ax1 = np.atleast_1d(axes)[i]
            
            #::: mark the TESS sectors      
            orb0 = Time(df['Start of Orbit'].loc[line]).jd
            orb1 = Time(df['End of Orbit'].loc[line]).jd
            orb2 = Time(df['Start of Orbit'].loc[line+1]).jd
            orb3 = Time(df['End of Orbit'].loc[line+1]).jd
            
            #::: add text and vspans
            ax1.text(0.98, 0.95, 'Sector '+str(s), ha='right', va='top', transform=ax1.transAxes, zorder=15)
            ax1.axvspan(orb0,orb1,color='b',alpha=0.1,zorder=-11)
            ax1.axvspan(orb2,orb3,color='b',alpha=0.1,zorder=-11)
            
            #::: set the axis xlim and labels properly
            ax1.set(xlim=[orb0-0.5, orb3+0.5])    
            _set_axes(ax1, time, y, labels={'show_x': (i==N-1)})    
        
    #::: iterate over all existing, set-up axes and plot
    for i, ax1 in enumerate(np.atleast_1d(axes)):
        
        #::: check if there is data to be plotted on this axis
        ind = np.where( (time > ax1.get_xlim()[0]) & (time <= ax1.get_xlim()[1]) )[0]
        
        #::: only proceed if there is actually data in this range
        if len(ind)>0:
            
            #::: plot (check if it should be a scatter or errorbar plot)
            _plot1(ax1, ind, time, y, yerr, mask_upper, mask_lower, clip, **kwargs)
                
    #::: return
    return axes
    


def tessplot_csv(fname, **kwargs):
    '''
    Wrapper around tessplot to plot straight from a csv file.
    See tessplot() for details.
    '''
    time, y, yerr = read_csv(fname)[0:3]
    return tessplot(time, y, yerr, **kwargs)   



def _initialize(time, y, yerr, ax, clip, kwargs):
    '''
    Prepares the data and figure. 
    '''
    #::: set default kwargs (via the pythonic spread operator)
    kwargs = {'color': 'b', 
              'marker': '.', 
              'ls': 'none', 
              'ms': 2,
              'rasterized': True,
              **kwargs}
    kwargs = {'ecolor': _get_errorbar_color(kwargs['color']),
              **kwargs}
        
    #::: clean up the data
    time, y, yerr = _clean_up(time, y, yerr)
    
    #::: clip the data (optional)
    if clip:
        y, mask_lower, mask_upper = _clip_helper(time, y)
    else:
        mask_lower, mask_upper = None, None
        
    #::: prepare the figure (optional)
    if ax is None: 
        fig, ax = plt.subplots(figsize=(12,3), tight_layout=True)
        
    #::: return
    return time, y, yerr, ax, kwargs, mask_lower, mask_upper



def _clean_up(time, y, yerr=None):    
    '''
    Just makes sure the input is in a format that can be handled.
    '''
    #::: convert to numpy arrays
    time = np.array(time)    
    y = np.array(y)    
    if yerr is not None: 
        yerr = np.array(yerr)    
    
    #::: return
    return time, y, yerr
        
        
        
def _clip_helper(time, y):
    '''
    Helps with clipping the data.
    '''
    y, _, mask_upper0, mask_lower0 = sigma_clip(time, y, low=4, high=4, return_mask=True)
    y, _, mask_upper, mask_lower = slide_clip(time, y, window_length=1, low=4, high=4, return_mask=True)
    mask_upper = np.array(mask_upper0 | mask_upper, dtype=float)
    mask_upper[mask_upper==0] = np.nan
    mask_lower = np.array(mask_lower0 | mask_lower, dtype=float)
    mask_lower[mask_lower==0] = np.nan
    
    return y, mask_lower, mask_upper



def _set_yticklabels(ax):
    ax.yaxis.set_major_locator(plt.MaxNLocator(3))



def _guess_ylabels(y):
    '''
    Sets the y and xlabels automatically by guessing whether it is a lightcurve, 
    RV, or a residuals time series.
    '''
    if (np.nanpercentile(y,5) > 0.7) & (np.nanpercentile(y,95) < 1.3):
        return 'Relative flux', None
    elif (np.nanpercentile(y,5) > -0.02) & (np.nanpercentile(y,95) < 0.02):
        return 'Residuals', None
    else:
        return 'RV', 'km/s'
        
        

def _set_xticklabels(ax, time_offset):
    '''
    Amends axes' xticklabels (rather than changing the value of "time" that we plot.)
    This is so that we do not cause complications if we want to reuse the same axis and plot something else on here.
    All subsequent plots can thus use the original time stamps and their data will be drawn correctly.
    '''
    def mynotation(time, pos):
        return '{:g}'.format(time - time_offset)
    MyFormatter = FuncFormatter(mynotation)
    ax.xaxis.set_major_formatter(MyFormatter)
    return ax



def _guess_xlabels(time):
    '''
    Guess the xlabels automatically between 'days' and 'BJD_TDB'.
    '''
    jd_moon_landing = 2440422.5
    if np.min(time) < jd_moon_landing:
        return 'Time', 'days'
    else:
        return 'Time', 'BJD$_\mathrm{TDB}$'
    


def _set_axes(ax, time, y, labels={}):
    '''
    Sets the axes ticks and labels automatically.
    '''
    #::: default labels and automatic guesses, unless overwritten by user-input
    labels = {'x': _guess_xlabels(time)[0],
              'x_unit': _guess_xlabels(time)[1],
              'y': _guess_ylabels(y)[0],
              'y_unit': _guess_ylabels(y)[1],
              'show_x': True,
              'show_y': True,
              **labels}
    
    #::: compute time offset (in steps of 1,000 days)
    time_offset = 1000*int(1.*np.min(time)/1e3)
    
    #::: handle the ylabels
    _set_yticklabels(ax)
    if labels['show_y']:
        if labels['y_unit'] is None:
            ax.set_ylabel( labels['y'] )
        else:
            ax.set_ylabel( labels['y'] + ' (' + labels['y_unit'] +')' )
        
    #::: handle the xlabels
    _set_xticklabels(ax, time_offset)
    if labels['show_x']:
        if labels['x_unit'] is None:
            ax.set_xlabel( labels['x'] + ' - ' + '{:,.0f}'.format(time_offset) )
        else:
            ax.set_xlabel( labels['x'] + ' (' + labels['x_unit'] + ' - ' + '{:,.0f}'.format(time_offset)+')' )
  
        
  
def _set_brokenaxes(bax, time, y, labels={}):
    '''
    Sets the axes ticks and labels automatically.
    '''
    #::: default labels and automatic guesses, unless overwritten by user-input
    labels = {'x': _guess_xlabels(time)[0],
              'x_unit': _guess_xlabels(time)[1],
              'y': _guess_ylabels(y)[0],
              'y_unit': _guess_ylabels(y)[1],
              'show_x': True,
              'show_y': True,
              **labels}
    
    #::: compute time offset (in steps of 1,000 days)
    time_offset = 1000*int(1.*np.min(time)/1e3)
    
    #::: handle the ylabels
    _set_yticklabels(bax.axs[0])
    if labels['show_y']:
        if labels['y_unit'] is None:
            bax.axs[0].set_ylabel( labels['y'] )
        else:
            bax.axs[0].set_ylabel( labels['y'] + ' (' + labels['y_unit'] +')' )
        
    #::: handle the xlabels
    for ax in bax.axs:
        _set_xticklabels(ax, time_offset)
    if labels['show_x']:
        if labels['x_unit'] is None:
            bax.set_xlabel( labels['x'] + ' - ' + '{:,.0f}'.format(time_offset) )
        else:
            bax.set_xlabel( labels['x'] + '(' + labels['x_unit'] + ' - ' + '{:,.0f}'.format(time_offset)+')' )
        
        

def _scale_lightness(rgb, scale_l):
    '''
    Returns a lightened version of any rgb color 
    '''
    # convert rgb to hls
    h, l, s = colorsys.rgb_to_hls(*rgb)
    # manipulate h, l, s values and return as rgb
    return colorsys.hls_to_rgb(h, min(1, l * scale_l), s = s)



def _get_errorbar_color(color):
    '''
    Converts a color string to rgb values and then returns a lighter version
    '''
    try:
        rgb = ColorConverter.to_rgb(color)
        return _scale_lightness(rgb, 1.75)
    except:
        return None



def _none_slice(x, ind):
    '''
    Tries to slice an array and skips None instances
    '''
    if x is None:
        return None
    else:
        return x[ind]

    

def _plot1(ax1, ind, time, y, yerr, mask_upper, mask_lower, clip, is_bax=False, **kwargs):
    '''
    Plots all the stuff onto one given axis
    '''
    #::: select the right data
    time1 = time[ind]
    y1 = y[ind]
    yerr1 = _none_slice(yerr, ind)
    mask_upper1 = _none_slice(mask_upper, ind)
    mask_lower1 = _none_slice(mask_lower, ind)
    
    #::: exception for brokenaxes
    if is_bax:
        ax1b = ax1.axs[-1]
    else: 
        ax1b = ax1
    
    #::: decide if it is an errorbar or scatter plot
    if len(kwargs['color']) < len(time):
        ax1.errorbar(time1, y1, yerr=yerr1, **kwargs)
    else:
        kwargs1 = kwargs.copy()
        color1 = kwargs['color'][ind]
        kwargs1.pop('color')
        kwargs1.pop('ecolor')
        kwargs1.pop('ls')
        kwargs1.pop('ms')
        kwargs1 = {'vmin': np.nanmin(kwargs['color']),
                   'vmax': np.nanmax(kwargs['color']),
                   **kwargs1} #this is important when breaking up the data in different axes
        sc = ax1.scatter(time1, y1, c=color1, **kwargs1)
        divider = make_axes_locatable(ax1b)
        cax = divider.append_axes('right', size='2%', pad=0.05)
        if is_bax: sc = sc[0]
        plt.colorbar(sc, cax=cax, orientation='vertical', label='Prob.') 
        #for now, the colorbar label is frozen to "Prob."; you can always overwrite it later
            
    #::: indicate clipped data (optional)
    if clip:
        ax1.plot(time1*mask_upper1, ax1b.get_ylim()[1]*mask_upper1, marker='^', ls='none', color='orange', ms=10, zorder=11)
        ax1.plot(time1*mask_lower1, ax1b.get_ylim()[0]*mask_lower1, marker='v', ls='none', color='orange', ms=10, zorder=11)





#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 22 10:47:38 2018

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
from .to_precision import std_notation




def round_to_2(x):
    if x==0:
        return x
    else:
        return round(x, -int(np.floor(np.log10(np.abs(x))))+1)

def round_to_reference(x, y):
    return round(x, -int(np.floor(np.log10(np.abs(y))))+1)

def str_digits(y):
    if np.abs(y)<1: return -int(np.floor(np.log10(np.abs(y))))+1
    else: return int(np.floor(np.log10(np.abs(y))))+1
    
def extra_digits(x,y):
    try:
        return int(np.floor(np.log10(np.abs(x)))) - int(np.floor(np.log10(np.abs(y))))
    except:
        return 0
    
def round_tex(x, err_low, err_up, mode=None):
    if np.isnan(x):
        return 'NaN'
    y = np.min(( np.abs(err_low), np.abs(err_up) ))
    digs = extra_digits(x,y) + 2
    if np.abs(err_low-err_up)/np.mean([err_low,err_up]) > 0.05:
        txt = std_notation(x,digs) + '_{-' + std_notation(err_low,2) + '}^{+' + std_notation(err_up,2) + '}'
    else:
        txt = std_notation(x,digs) + '\\pm' + std_notation( np.max(( np.abs(err_low), np.abs(err_up) )) ,2)
    if mode is None:
        return txt
    else:
        return txt, std_notation(mode,digs)

def round_txt_separately(x, err_low, err_up):
    if np.isnan(x):
        return 'NaN'
    y = np.min(( np.abs(err_low), np.abs(err_up) ))
    digs = extra_digits(x,y) + 2
    txt1 = std_notation(x,digs) 
    txt2 = std_notation(err_low,2) 
    txt3 = std_notation(err_up,2)
    return txt1, txt2, txt3


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr 15 11:15:29 2020

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from astropy.timeseries import LombScargle
from statsmodels.graphics.tsaplots import plot_acf

#::: my modules
from ..exoworlds_rdx.lightcurves.lightcurve_tools import plot_phase_folded_lightcurve                                                                                   
from allesfitter.time_series import clean, sigma_clip, slide_clip
                                                                               
#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




###############################################################################
#::: run a periodogram via astropy to get the dominant period and FAP
###############################################################################
def estimate_period(time, y, y_err, clip=True, plot=True, **kwargs):
    """
    Run a Lomb-Scargle Periodogram to find periodic signals. It's recommended 
    to use the allesfitter.time_series functions sigma_clip and slide_clip beforehand.

    Parameters
    ----------
    time : array of float
        e.g. time array (usually in days)
    y : array of float
        e.g. flux or RV array (usually as normalized flux or RV in km/s)
    yerr : array of float
        e.g. flux or RV error array (usually as normalized flux or RV in km/s)
    clip : bool, optional
        Automatically clip the input data with sigma_clip(low=4, high=4)
        and slide_clip(window_length=1, low=4, high=4). The default is True.
    plot : bool, optional
        To plot or not, that is the question. The default is False.
    **kwargs : collection of keyword arguments
        Any keyword arguments will be passed onto the astropy periodogram class.

    Returns
    -------
    best_period : float
        The best period found.
    FAP : float
        The false alarm probability for the best period.
    fig : matplotlib.figure object, optional
        The summary figure. Only returned if plot is True.
    """
    
    #==========================================================================
    #::: clean the inputs
    #==========================================================================
    time, y, y_err = clean(time, y, y_err)
    plot_bool = plot
    
    if clip:
        y = sigma_clip(time, y, low=4, high=4)
        y = slide_clip(time, y, window_length=1, low=4, high=4)
        time, y, y_err = clean(time, y, y_err)
        
    
    #==========================================================================
    #::: handle inputs
    #==========================================================================
    cadence = np.nanmedian(np.diff(time))
    if kwargs is None: kwargs = {}
    if 'minperiod' not in kwargs: kwargs['minperiod'] = 10. * cadence
    if 'maxperiod' not in kwargs: kwargs['maxperiod'] = time[-1]-time[0]
    minfreq = 1./kwargs['maxperiod']
    maxfreq = 1./kwargs['minperiod']
    
    
    #==========================================================================
    #::: now do the periodogram
    #==========================================================================
    ls = LombScargle(time, y) #Analyze our dates and s-index data using the AstroPy Lomb Scargle module
    frequency, power = ls.autopower(minimum_frequency=minfreq, maximum_frequency=maxfreq) #Determine the LS periodogram
    best_power = np.nanmax(power)
    best_frequency = frequency[np.argmax(power)]
    best_period = 1./best_frequency
    FAP=ls.false_alarm_probability(best_power) #Calculate the FAP for the highest peak in the power array  
    
    
    #==========================================================================
    #::: plot
    #==========================================================================
    def plot(): 
        
        peak_loc=round(float(1./best_frequency),2) 
        FAP_probabilities = [0.5, 0.1, 0.01] #Enter FAP values you want to determine
        FAP_levels=ls.false_alarm_level(FAP_probabilities) #Get corresponding LS Power values
        
        fig, axes = plt.subplots(4, 1, figsize=[10,15], tight_layout=True)
        
        #::: plot the periodogram
        ax = axes[0]        
        ax.semilogx(1./frequency,power,color='b')  
        ax.plot(peak_loc, best_power, marker='d', markersize=12, color='r')                
        ax.text(peak_loc*1.2,best_power*0.95,'Peak Period: '+str(peak_loc)+' days')
        ax.text(peak_loc*1.2,best_power*0.85,'FAP: '+str(FAP))    
        ax.hlines(FAP_levels, kwargs['minperiod'], kwargs['maxperiod'], color='grey', lw=1)    
        ax.text(kwargs['maxperiod'], FAP_levels[0],'0.5% FAP ', ha='right')     
        ax.text(kwargs['maxperiod'], FAP_levels[1],'0.1% FAP ', ha='right')
        ax.text(kwargs['maxperiod'], FAP_levels[2],'0.01% FAP ', ha='right')
        ax.set(xlabel='Period (days)', ylabel='L-S power')   
        ax.tick_params(axis='both',which='major')    
        
        #::: plot the phase-folded data
        ax = axes[1]
        plot_phase_folded_lightcurve(time, y, period=1./best_frequency, epoch=0, ax=ax)
        ax.set(ylim=[np.nanmin(y), np.nanmax(y)], ylabel='Data (clipped; phased)')
        
        #::: plot the phase-folded data, zoomed
        ax = axes[2]
        plot_phase_folded_lightcurve(time, y, period=1./best_frequency, epoch=0, ax=ax)
        ax.set(ylabel='Data (clipped; phased; y-zoom)')
        
        #::: plot the autocorrelation of the data
        ax = axes[3]
        plot_acf(pd.Series(y, index=time), ax=ax, lags=np.linspace(start=1,stop=2*best_period/cadence,num=100,dtype=int))
        ax.set(xlabel='Lag', ylabel='Autocorrelation', title='')
    
        return fig
    
    
    #==========================================================================
    #::: return
    #==========================================================================
    if plot_bool:
        fig = plot()
        return best_period, FAP, fig
    else:
        return best_period, FAP



# ###############################################################################
# #::: run a periodogram via astropy to get the dominant period and FAP
# ###############################################################################
def estimate_period_old(time, y, y_err, periodogram_kwargs=None, astropy_kwargs=None, wotan_kwargs=None, options=None):
    '''
    Parameters
    ----------
    time : TYPE
        DESCRIPTION.
    y : TYPE
        DESCRIPTION.
    y_err : TYPE
        DESCRIPTION.
    periodogram_kwargs : TYPE, optional
        DESCRIPTION. The default is None.
    astropy_kwargs : TYPE, optional
        DESCRIPTION. The default is None.
    wotan_kwargs : TYPE, optional
        DESCRIPTION. The default is None.
    options : None or dictionary, optional
        The default is None, which will evaluate to:
            options = {}
            options['show_plot'] = True #show a plot in the terminal?
            options['save_plot'] = True #save a plot?
            options['fname_plot'] = 'periodogram' #filenmae of the plot
            options['outdir'] = '.' #output directory for the plot
        If a dictionary is given, it may contain and overwrite all these keys.

    Returns
    -------
    None.
    '''  
    
    #==========================================================================
    #::: handle inputs
    #==========================================================================
    cadence = np.nanmedian(np.diff(time))
        
    if periodogram_kwargs is None: periodogram_kwargs = {}
    if 'minperiod' not in periodogram_kwargs: periodogram_kwargs['minperiod'] = 10. * cadence
    if 'maxperiod' not in periodogram_kwargs: periodogram_kwargs['maxperiod'] = time[-1]-time[0]
    
    if astropy_kwargs is None: astropy_kwargs = {}
    if 'sigma' not in astropy_kwargs: astropy_kwargs['sigma'] = 5
    
    if wotan_kwargs is None: wotan_kwargs = {}
    if 'slide_clip' not in wotan_kwargs: wotan_kwargs['slide_clip'] = {}
    if 'window_length' not in wotan_kwargs['slide_clip']: wotan_kwargs['slide_clip']['window_length'] = 1.
    if 'low' not in wotan_kwargs['slide_clip']: wotan_kwargs['slide_clip']['low'] = 5
    if 'high' not in wotan_kwargs['slide_clip']: wotan_kwargs['slide_clip']['high'] = 5

    if options is None: options = {}
    if 'show_plot' not in options: options['show_plot'] = False
    if 'save_plot' not in options: options['save_plot'] = False
    if 'return_plot' not in options: options['return_plot'] = False
    if 'fname_plot' not in options: options['fname_plot'] = 'periodogram'
    if 'outdir' not in options: options['outdir'] = '.'
    
    minfreq = 1./periodogram_kwargs['maxperiod']
    maxfreq = 1./periodogram_kwargs['minperiod']
    
    
    #==========================================================================
    #::: first, a global 5 sigma clip
    #==========================================================================
    ff = sigma_clip(time, np.ma.masked_invalid(y), low=astropy_kwargs['sigma'], high=astropy_kwargs['sigma']) #astropy wants masked arrays
    # ff = np.array(ff.filled(np.nan)) #use NaN instead of masked arrays, because masked arrays drive me crazy
    
    
    #==========================================================================
    #::: fast slide clip (1 day, 5 sigma) [replaces Wotan's slow slide clip]
    #==========================================================================
    try: ff = slide_clip(time, ff, **wotan_kwargs['slide_clip'])
    except: print('Fast slide clip failed and was skipped.')     
    
    
    #==========================================================================
    #::: now do the periodogram
    #==========================================================================
    ind_notnan = np.where(~np.isnan(time*ff*y_err))
    ls = LombScargle(time[ind_notnan], ff[ind_notnan])  #Analyze our dates and s-index data using the AstroPy Lomb Scargle module
    frequency, power = ls.autopower(minimum_frequency=minfreq, maximum_frequency=maxfreq)  #Determine the LS periodogram
    best_power = np.nanmax(power)
    best_frequency = frequency[np.argmax(power)]
    FAP=ls.false_alarm_probability(best_power)                  #Calculate the FAP for the highest peak in the power array  
    
    
    #==========================================================================
    #::: plots
    #==========================================================================
    if options['show_plot'] or options['save_plot'] or options['return_plot']: 
        
        peak_loc=round(float(1./best_frequency),2) 
        FAP_probabilities = [0.5, 0.1, 0.01]                         #Enter FAP values you want to determine
        FAP_levels=ls.false_alarm_level(FAP_probabilities)           #Get corresponding LS Power values
        
        fig, axes = plt.subplots(4, 1, figsize=[10,15], tight_layout=True)  
        # axes = np.atleast_1d(axes)
        
        # ax = axes[0]   
        # ind_clipped = np.where(np.isnan(ff))[0]
        # ax.plot(time[ind_clipped], flux[ind_clipped], 'r.', rasterized=True)
        # ax.plot(time, ff, 'b.', rasterized=True)
        # ax.set(xlabel='Time (BJD)', ylabel='Flux')
        
        # ax = axes[1]       
        # ax.plot(time, ff, 'b.', rasterized=True)
        # ax.set(xlabel='Time (BJD)', ylabel='Flux (clipped)')
        
        ax = axes[0]        
        ax.semilogx(1./frequency,power,color='b')  
        ax.plot(peak_loc, best_power, marker='d', markersize=12, color='r')                
        ax.text(peak_loc*1.2,best_power*0.95,'Peak Period: '+str(peak_loc)+' days')
        ax.text(peak_loc*1.2,best_power*0.85,'FAP: '+str(FAP))    
        ax.hlines(FAP_levels, periodogram_kwargs['minperiod'], periodogram_kwargs['maxperiod'], color='grey', lw=1)    
        ax.text(periodogram_kwargs['maxperiod'], FAP_levels[0],'0.5% FAP ', ha='right')     
        ax.text(periodogram_kwargs['maxperiod'], FAP_levels[1],'0.1% FAP ', ha='right')
        ax.text(periodogram_kwargs['maxperiod'], FAP_levels[2],'0.01% FAP ', ha='right')
        ax.set(xlabel='Period (days)', ylabel='L-S power')   
        ax.tick_params(axis='both',which='major')    
#        ax.text(peak_loc*1.2,best_power*0.75,'std_old:'+str(std_old*1e3)[0:4]+' --> '+'std_new:'+str(std_new*1e3)[0:4])    
        
        ax = axes[1]
        plot_phase_folded_lightcurve(time, ff, period=1./best_frequency, epoch=0, ax=ax)
        ax.set(ylim=[np.nanmin(ff), np.nanmax(ff)], ylabel='Data (clipped; phased)')
        
        ax = axes[2]
        plot_phase_folded_lightcurve(time, ff, period=1./best_frequency, epoch=0, ax=ax)
        ax.set(ylabel='Data (clipped; phased; y-zoom)')
        
        #::: plot the autocorrelation of the data
        ax = axes[3]
        plot_acf(pd.Series(ff[ind_notnan], index=time[ind_notnan]), ax=ax, lags=np.linspace(start=1,stop=2000,num=100,dtype=int))
        ax.set(xlabel='Lag', ylabel='Autocorrelation', title='')
    
        if options['save_plot']:
            if not os.path.exists(options['outdir']): os.makedirs(options['outdir'])
            fig.savefig(os.path.join(options['outdir'],options['fname_plot']+'.pdf'), bbox_inches='tight')
        if options['show_plot']:
            plt.show(fig)
        else:
            plt.close(fig)
    
    if options['return_plot'] is True:     
        return 1./best_frequency, FAP, axes
    else:
        return 1./best_frequency, FAP  



###############################################################################
#::: estimate a good window length for spline knots, running median etc.
###############################################################################
# def estimate_window_length(time, flux, flux_err, periodogram_kwargs=None, wotan_kwargs=None, options=None):
#     window_length_min = 12./24. #at least 12h to not destroy planets
#     window_length_max = 1. #at most 1 day
#     cadence = np.median(np.diff(time))
#     best_period, FAP = estimate_period(time, flux, flux_err)
    
#     if best_period < 100.*cadence: 
#         window_length = best_period/10.
#     return np.min()
    
#         return best_period/10.
#     else:
#         return None
#         warnings.warn('Returning None. Best period was found to be', best_period*24*60, 'min., but cadence is only', cadence*24*60, 'min.')
        
        

###############################################################################
#::: remove periodic trends
###############################################################################
# def estimate_trend(time, flux, flux_err):
    
#     iterations = 3 
#     wotan_kwargs = {'slide_clip':{}, 'flatten':{}}
#     wotan_kwargs['slide_clip']['window_length'] = 1.
#     wotan_kwargs['slide_clip']['low'] = 3.
#     wotan_kwargs['slide_clip']['high'] = 3.
#     wotan_kwargs['flatten']['method'] = 'rspline'
#     wotan_kwargs['flatten']['window_length'] = None
#     wotan_kwargs['flatten']['break_tolerance'] = 0.5
    
#     trend = np.ones_like(time)
    
#     #::: global sigma clipping
#     flux = sigma_clip(flux, sigma_upper=3, sigma_lower=20)
    
#     #::: 1 day slide clip
#     flux = slide_clip(time, flux, **wotan_kwargs['slide_clip'])
    
#     for i in range(iterations):
#         wotan_kwargs['flatten']['window_length'] = estimate_window_length(time, flux, flux_err)
#         print(wotan_kwargs['flatten']['window_length']*10)
#         if wotan_kwargs['flatten']['window_length'] is not None:
#             flux, trend1 = flatten(time, flux, return_trend=True, **wotan_kwargs['flatten'])
#             trend += (trend1 - 1.)
#         plt.figure()
#         plt.plot(time, flux)
#         plt.title(i)
        
         
#     return trend
    
    
    


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 30 12:14:51 2018

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
import matplotlib.pyplot as plt

#::: allesfitter modules
from .. import get_mcmc_posterior_samples, get_ns_posterior_samples, get_labels

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})




def mcmc_plot_violins(datadirs, labels, key):
    '''
    Inputs:
    -------
    
    Outputs:
    -------
    violin plots comparing the parameters of different runs
    
    
    Example:
    --------
    datadirs = ['allesfit_global/allesfit_wouttess_ns',
                'allesfit_global/allesfit_onlytess_ns',
                'allesfit_global/allesfit_withtess_ns',
                ]
    labels = ['w/o TESS',
              'only TESS',
              'w/ TESS',
              ]
    ns_plot_violins(datadirs, labels)
    '''
    return plot_violins(datadirs, labels, key, 'mcmc')
   
   
   

def ns_plot_violins(datadirs, labels, key):
    '''
    Inputs:
    -------
    
    Outputs:
    -------
    violin plots comparing the parameters of different runs
    
    
    Example:
    --------
    datadirs = ['allesfit_global/allesfit_wouttess_ns',
                'allesfit_global/allesfit_onlytess_ns',
                'allesfit_global/allesfit_withtess_ns',
                ]
    labels = ['w/o TESS',
              'only TESS',
              'w/ TESS',
              ]
    ns_plot_violins(datadirs, labels)
    '''
    return plot_violins(datadirs, labels, key, 'ns')
   
  
    

def plot_violins(datadirs, labels, key, mode):  
    
    all_params = {}
    all_paramslabels = {}
    for datadir, label in zip(datadirs, labels):
        if mode=='mcmc':        
            all_params[label] = get_mcmc_posterior_samples(datadir, as_type='dic')
        elif mode=='ns':
            all_params[label] = get_ns_posterior_samples(datadir, as_type='dic')
        all_paramslabels[label] = get_labels(datadir, as_type='dic')
        
    xsize= max(6, 1.5*len(labels))
    fig, ax = plt.subplots(figsize=(xsize,4))
    violinlist = []
    for l in labels:
        if key in all_params[l]:
            violinlist.append( all_params[l][key] )
        else:
            violinlist.append( np.ones(10)*np.nan )
    positionlist = range(len(labels))
    ax.violinplot(violinlist, positions=positionlist, showmedians=True, showextrema=False)
    ax.set_xticks(positionlist)
    ax.set_xticklabels(labels)
    ax.set_ylabel(all_paramslabels[labels[0]][key])
    plt.tight_layout()
    return fig, ax


#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Oct 11 19:51:37 2018

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np




###############################################################################
#::: expand flags
###############################################################################
def expand_flags(flag, n=4):
    maskleft = np.where(flag>0)[0]
    flag_new = np.zeros(len(flag),dtype=bool)
    for i in range(-n,n+1):
        thismask = maskleft+i
        for t in thismask:
            if t<len(flag_new): flag_new[t] = 1
    return flag_new



# -*- coding: utf-8 -*-
"""
Created on Tue Apr  5 15:07:00 2016

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import numpy as np
from astropy.coordinates import SkyCoord
import astropy.units as u
import pandas as pd
from pprint import pprint



###########################################################################
#::: Statistics and time series helpers
###########################################################################

        
def medsig(a):
    '''Compute median and MAD-estimated scatter of array a'''
    med = np.nanmedian(a)
    sig = 1.48 * np.nanmedian(abs(a-med))
    return med, sig   
    
  
  
def running_mean(x, N):
    x[np.isnan(x)] = 0. #reset NAN to 0 to calculate the cumulative sum; mimics the 'pandas' behavior
    cumsum = np.cumsum(np.insert(x, 0., 0.)) 
    return 1.*(cumsum[N:] - cumsum[:-N]) / N 
    


# 'running_median' DOES NOT AGREE WITH THE PANDAS IMPLEMENTATION 'running_median_pandas'
#def running_median(x, N):
#    return np.array(list(RunningMedian(N, x)))


    
def running_mean_pandas(x, N):
    ts = pd.Series(x).rolling(window=N, center=False).mean()
    return ts[~np.isnan(ts)].as_matrix()



def running_median_pandas(x, N):
    ts = pd.Series(x).rolling(window=N, center=False).median()
    return ts[~np.isnan(ts)].as_matrix()


    
def mask_ranges(x, x_min, x_max):
    """"
    Crop out values and indices out of an array x for multiple given ranges x_min to x_max.
    
    Input:
    x: array, 
    x_min: lower limits of the ranges
    x_max: upper limits of the ranges
    
    Output:
    
    
    Example:
    x = np.arange(200)    
    x_min = [5, 25, 90]
    x_max = [10, 35, 110]
    """

    mask = np.zeros(len(x), dtype=bool)
    for i in range(len(x_min)): 
        mask = mask | ((x >= x_min[i]) & (x <= x_max[i]))
    ind_mask = np.arange(len(mask))[mask]
    
    return x[mask], ind_mask, mask 




###########################################################################
#::: Text formatting for tables and plots
###########################################################################



def mystr(x,digits=0):
    if np.isnan(x): return '.'
    elif digits==0: return str(int(round(x,digits)))
    else: return str(round(x,digits))



def format_2sigdigits(x1, x2, x3, nmax=3):
    n = int( np.max( [ -np.floor(np.log10(np.abs(x))) for x in [x1, x2, x3] ] ) + 1 )
    scaling = 0
    extra = None
    if n > nmax:
        scaling = n-1
        n = 1
        extra = r"\cdot 10^{" + str(-scaling) + r"}"
    return str(round(x1*10**scaling, n)).ljust(n+2, '0'), str(round(x2*10**scaling, n)).ljust(n+2, '0'), str(round(x3*10**scaling, n)).ljust(n+2, '0'), extra



def deg2hmsdms(ra, dec):
    c = SkyCoord(ra=ra * u.degree, dec=dec * u.degree, frame='icrs')
    return c.to_string('hmsdms', precision=2, sep=':')



def format_latex(x1, x2, x3, nmax=3):
    r, l, u, extra = format_2sigdigits(x1, x2, x3, nmax)
    if l==u:
        core = r + r"\pm" + l
    else:
        core = r + r"^{+" + l + r"}_{-" + u + r"}"

    if extra is None:
        return r"$" + core + r"$"
    else:
        return r"$" + '(' + core + ')' + extra + r"$"



###########################################################################
#::: Dictionaries and tables
###########################################################################


def merge_two_dicts(x, y):
    '''Given two dicts, merge them into a new dict as a shallow copy.'''
    z = x.copy()
    z.update(y)
    return z
        
        
def table_view(dic):
    from astropy.table import Table 
    dic_table = {}
    subkeys = ['OBJ_ID', 'SYSREM_FLUX3_median', 'PERIOD', 'DEPTH', 'WIDTH', 'NUM_TRANSITS']
    for key in subkeys:
            dic_table[key] = dic[key]
    dic_table = Table(dic_table)
    dic_table = dic_table[subkeys]
    pprint(dic_table)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Apr  8 18:34:28 2020

@author:
Maximilian N. Günther
MIT Kavli Institute for Astrophysics and Space Research, 
Massachusetts Institute of Technology,
77 Massachusetts Avenue,
Cambridge, MA 02109, 
USA
Email: maxgue@mit.edu
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt

#::: my modules
# from .translator import translate




###############################################################################
#::: get the defaults
###############################################################################
def get_hot_jupiter_params():
    default_params = {}
    default_params['b_rr'] = 0.1
    default_params['b_rsuma'] = 0.1
    default_params['b_cosi'] = 0.
    default_params['b_epoch'] = 0.
    default_params['b_period'] = 1.
    default_params['b_K'] = 0.
    default_params['b_f_c'] = 0.
    default_params['b_f_s'] = 0.
    default_params['dil_Leonardo'] = 0.
    default_params['b_sbratio_Leonardo'] = 0.
    default_params['host_ldc_q1_Leonardo'] = 0.44502241000000003 #in q-space; quadratic limb darkening for a solar-type star
    default_params['host_ldc_q2_Leonardo'] = 0.3600659571278669 #in q-space; quadratic limb darkening for a solar-type star
    default_params['host_gdc_Leonardo'] = None
    default_params['b_gdc_Leonardo'] = None
    return default_params



def get_hot_jupiter_params_host():
    default_params_host = {'R_host':1,
                           'M_host':1,
                           'Teff_host':5700}
    return default_params_host
                      


def get_hot_jupiter_settings():
    default_settings = {'companions_phot':['b'],
                        'companions_rv':[],
                        'inst_phot':['Leonardo'],
                        'inst_rv':[],
                        'fit_ttvs':False,
                        'host_ld_law_Leonardo':'quad',
                        'b_ld_law_Leonardo':None}
    return default_settings



###############################################################################
#::: fill missing params with defaults
###############################################################################
def fill_params(params=None, settings=None):
    
    if params is None: 
        params = {}
    
    if settings is None:
        raise ValueError('Must input settings.')
    else:
        companions_all = list(np.unique(settings['companions_phot'] + settings['companions_rv']))
        inst_all = list(np.unique(settings['inst_phot'] + settings['inst_rv']))
    
    for companion in companions_all:
        for inst in inst_all:
    
            if 'dil_'+inst not in params:
                params['dil_'+inst] = 0.
            
            if companion+'_rr' not in params:
                params[companion+'_rr'] = None
                
            if companion+'_rsuma' not in params:
                params[companion+'_rsuma'] = None
                
            if companion+'_cosi' not in params:
                params[companion+'_cosi'] = 0.
                
            if companion+'_epoch' not in params:
                params[companion+'_epoch'] = None
                
            if companion+'_period' not in params:
                params[companion+'_period'] = None
                
            if companion+'_sbratio_'+inst not in params:
                params[companion+'_sbratio_'+inst] = 0.               
                
            if companion+'_a' not in params:
                params[companion+'_a'] = None
                
            if companion+'_q' not in params:
                params[companion+'_q'] = 1.
                
            if companion+'_K' not in params:
                params[companion+'_K'] = 0.
            
            if companion+'_f_c' not in params:
                params[companion+'_f_c'] = 0.
                
            if companion+'_f_s' not in params:
                params[companion+'_f_s'] = 0.
                
            if 'host_ldc_'+inst not in params:
                params['host_ldc_'+inst] = None
                
            if companion+'_ldc_'+inst not in params:
                params[companion+'_ldc_'+inst] = None
                
            if 'host_gdc_'+inst not in params:
                params['host_gdc_'+inst] = None
                
            if companion+'_gdc_'+inst not in params:
                params[companion+'_gdc_'+inst] = None
                
            if 'didt_'+inst not in params:
                params['didt_'+inst] = None
                
            if 'domdt_'+inst not in params:
                params['domdt_'+inst] = None
                
            if 'host_rotfac_'+inst not in params:
                params['host_rotfac_'+inst] = 1.
                
            if companion+'_rotfac_'+inst not in params:
                params[companion+'_rotfac_'+inst] = 1.
                
            if 'host_hf_'+inst not in params:
                params['host_hf_'+inst] = 1.5
                
            if companion+'_hf_'+inst not in params:
                params[companion+'_hf_'+inst] = 1.5
                
            if 'host_bfac_'+inst not in params:
                params['host_bfac_'+inst] = None
                
            if companion+'_bfac_'+inst not in params:
                params[companion+'_bfac_'+inst] = None
                
            if 'host_heat_'+inst not in params:
                params['host_heat_'+inst] = None
                
            if companion+'_heat_'+inst not in params:
                params[companion+'_heat_'+inst] = None
                
            if 'host_lambda_'+inst not in params:
                params['host_lambda_'+inst] = None
                
            if companion+'_lambda_'+inst not in params:
                params[companion+'_lambda_'+inst] = None
                
            if 'host_vsini' not in params:
                params['host_vsini'] = None
                
            if companion+'_vsini' not in params:
                params[companion+'_vsini'] = None
                
            if 'host_spots_'+inst not in params:
                params['host_spots_'+inst] = None
                
            if companion+'_spots_'+inst not in params:
                params[companion+'_spots_'+inst] = None
                
            if companion+'_phase_curve_beaming_'+inst not in params: #in ppt
                params[companion+'_phase_curve_beaming_'+inst] = None
                
            if companion+'_phase_curve_atmospheric_'+inst not in params: #in ppt
                params[companion+'_phase_curve_atmospheric_'+inst] = None
                
            if companion+'_phase_curve_ellipsoidal_'+inst not in params: #in ppt
                params[companion+'_phase_curve_ellipsoidal_'+inst] = None
            
    return params



def fill_params_host(params_host=None):
    
    if params_host is None: 
        params_host = {}

    if 'R_host' not in params_host:
        params_host['R_host'] = None
        
    if 'M_host' not in params_host:
        params_host['M_host'] = None
        
    if 'Teff_host' not in params_host:
        params_host['Teff_host'] = None
        
    return params_host
        


def fill_settings(settings=None):
    
    if settings is None: 
        settings = {}

    if 'companions_phot' not in settings:
        settings['companions_phot'] = []
        
    if 'companions_rv' not in settings:
        settings['companions_rv'] = []
        
    if 'inst_phot' not in settings:
        settings['inst_phot'] = []
        
    if 'inst_rv' not in settings:
        settings['inst_rv'] = []
        
    if 'phase_curve' not in settings:
        settings['phase_curve'] = False
        
    if 'phase_curve_style' not in settings:
        settings['phase_curve_style'] = None
        
    if 'fit_ttvs' not in settings:
        settings['fit_ttvs'] = False
        
    if 'exact_grav' not in settings:
        settings['exact_grav'] = False
        
    if 'N_flares' not in settings:
        settings['N_flares'] = 0.
        
    if 'N_spots' not in settings:
        settings['N_spots'] = 0.
        
    companions_all = list(np.unique(settings['companions_phot'] + settings['companions_rv']))
    inst_all = list(np.unique(settings['inst_phot'] + settings['inst_rv']))
    
    for companion in companions_all:
        for inst in inst_all:
            
            if 'host_ld_law_'+inst not in settings:
                settings['host_ld_law_'+inst] = None
                
            if companion+'_ld_law_'+inst not in settings:
                settings[companion+'_ld_law_'+inst] = None
            
            if 'host_grid_'+inst not in settings:
                settings['host_grid_'+inst] = 'default'
                
            if companion+'_grid_'+inst not in settings:
                settings[companion+'_grid_'+inst] = 'default'
            
            if 'host_shape_'+inst not in settings:
                settings['host_shape_'+inst] = 'sphere'
                
            if companion+'_shape_'+inst not in settings:
                settings[companion+'_shape_'+inst] = 'sphere'
        
    return settings



        
# def get_empty_params_host():
#     default_params_host = {'R_host':None,
#                            'M_host':None,
#                            'Teff_host':None}
#     return default_params_host
                      


# def get_empty_settings():
#     default_settings = {'companions_phot':[''],
#                         'companions_rv':[''],
#                         'inst_phot':[''],
#                         'inst_rv':[''],
#                         'fit_ttvs':False,
#                         'ld':'quad'}
#     return default_settings



###############################################################################
#::: fill gaps in an exisiting dictionary with defaults
###############################################################################
# def fill_params(params, **kwargs):
#     if params is None: params = {}
#     default_params = get_default_params()
#     for key in default_params.keys():
#         if key not in params: 
#             params[key] = default_params[key]
#     for key in kwargs:
#         if key in params:
#             params[key] = kwargs[key]
#     return params



# def fill_params_host(params_host, **kwargs):
#     if params_host is None: params_host = {}
#     default_params_host = get_default_params_host()
#     for key in default_params_host.keys():
#         if key not in params_host: 
#             params_host[key] = default_params_host[key]
#     for key in kwargs:
#         if key in params_host:
#             params_host[key] = kwargs[key]
#     return params_host



# def fill_settings(settings, **kwargs):
#     if settings is None: settings = {}
#     default_settings = get_default_settings()
#     for key in default_settings.keys():
#         if key not in settings: 
#             settings[key] = default_settings[key]
#     for key in kwargs:
#         if key in settings:
#             settings[key] = kwargs[key]
#     return settings





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 14 17:55:39 2020

@author:
Dr. Maximilian N. Günther
European Space Agency (ESA)
European Space Research and Technology Centre (ESTEC)
Keplerlaan 1, 2201 AZ Noordwijk, The Netherlands
Email: maximilian.guenther@esa.int
GitHub: mnguenther
Twitter: m_n_guenther
Web: www.mnguenther.com
"""

from __future__ import print_function, division, absolute_import

#::: modules
import os, sys
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import warnings
from pprint import pprint
from datetime import datetime
from astropy import units as u
from astropy import constants as c
from astropy.stats import sigma_clip
from astropy.timeseries import BoxLeastSquares as bls
from ..exoworlds_rdx.lightcurves.index_transits import index_transits
# import time as timer
import contextlib

#::: specific modules
try:
    from wotan import flatten
except ImportError:
    pass

try:
    from transitleastsquares import transitleastsquares as tls
    from transitleastsquares import transit_mask, catalog_info
except ImportError:
    pass

#::: my modules
try:
    from exoworlds.tess import tessio
except:
    pass
from ..exoworlds_rdx.lightcurves.lightcurve_tools import plot_phase_folded_lightcurve, rebin_err  
from ..time_series import clean, slide_clip
from ..lightcurves import tessclean
from ..inout import write_json, write_csv
from ..plotting import fullplot, brokenplot, tessplot, monthplot


#::: plotting settings
import seaborn as sns
sns.set(context='paper', style='ticks', palette='deep', font='sans-serif', font_scale=1.5, color_codes=True)
sns.set_style({"xtick.direction": "in","ytick.direction": "in"})
sns.set_context(rc={'lines.markeredgewidth': 1})

    

###############################################################################
#::: print to logfile
###############################################################################
def logprint(*text, options=None):
    original = sys.stdout
    try:
        with open(os.path.join(options['outdir'],'logfile.log'), 'a' ) as f:
            sys.stdout = f
            print(*text)
    except OSError:
        pass #For unknown reasons, the combination of open() and os.path.join() does not work on some Windows versions
    sys.stdout = original
    
    
    
###############################################################################
#::: pretty-print to logfile
###############################################################################
def logpprint(*text, options=None):
    original = sys.stdout
    try:
        with open(os.path.join(options['outdir'],'logfile.log'), 'a' ) as f:
            sys.stdout = f
            pprint(*text)
    except OSError:
        pass #For unknown reasons, the combination of open() and os.path.join() does not work on some Windows versions
    sys.stdout = original


    
###############################################################################
#::: apply a mask (if wished so)
###############################################################################
def mask(time, flux, flux_err, period, duration, T0):
    intransit = transit_mask(time, period, duration, T0)
    time = time[~intransit]
    flux = flux[~intransit]
    if flux_err is not None:
        flux_err = flux_err[~intransit]
        time, flux, flux_err = clean(time, flux, flux_err)
    else:
        time, flux = clean(time, flux)
    return time, flux, flux_err
    


###############################################################################
#::: check for multiples of a value (e.g., of a period)
###############################################################################
def is_multiple_of(a, b, tolerance=0.05):
    a = np.float(a)
    b = np.float(b) 
    result = a % b
    return (abs(result/b) <= tolerance) or (abs((b-result)/b) <= tolerance)



###############################################################################
#::: BLS search on an input lightcurve
###############################################################################
def bls_search(time, flux, flux_err=None):
    if flux_err is None: 
        ind = np.where(~np.isnan(time*flux))
        time = np.array(time)[ind]
        flux = np.array(flux)[ind]
    else:
        ind = np.where(~np.isnan(time*flux*flux_err))
        time = np.array(time)[ind]
        flux = np.array(flux)[ind]
        flux_err = np.array(flux_err)[ind]
    print(time, flux)
    plt.figure()
    plt.plot(time, flux, 'b.')
    model = bls(time * u.day, flux, dy=flux_err)
    print(model)
    periodogram = model.autopower(0.05)
    plt.plot(periodogram.period, periodogram.power)  
    # max_power = np.argmax(periodogram.power)
    # stats = model.compute_stats(periodogram.period[max_power],
    #                             periodogram.duration[max_power],
    #                             periodogram.transit_time[max_power])
    # print(stats)
        
    
    
###############################################################################
#::: get TLS kwargs from TICv8
###############################################################################
def get_tls_kwargs_by_tic(tic_id, sigma=3, tls_kwargs=None, quiet=True):
    #mass comes first, radius comes second in the TLS source code for catalog_info()
    u, M_star, M_star_lerr, M_star_uerr, R_star, R_star_lerr, R_star_uerr = catalog_info(TIC_ID=int(tic_id))
    if not quiet:
        print('TICv8 info:')
        print('Quadratic limb darkening u_0, u_1', u[0], u[1])
        print('Stellar radius', R_star, '+', R_star_lerr, '-', R_star_uerr)
        print('Stellar mass', M_star, '+', M_star_lerr, '-', M_star_uerr)
        
    if tls_kwargs is None: tls_kwargs = {}
    tls_kwargs['R_star']=float(R_star)
    tls_kwargs['R_star_min']=R_star-sigma*R_star_lerr
    tls_kwargs['R_star_max']=R_star+sigma*R_star_uerr
    tls_kwargs['M_star']=float(M_star)
    tls_kwargs['M_star_min']=M_star-sigma*M_star_lerr
    tls_kwargs['M_star_max']=M_star+sigma*M_star_uerr
    tls_kwargs['u']=u
    
    if np.isnan(tls_kwargs['R_star']): 
        tls_kwargs['R_star'] = 1.
        warnings.warn("tls_kwargs: R_star was undefined in TICv8. Filling it with R_star=1.")
    if np.isnan(tls_kwargs['R_star_min']): 
        tls_kwargs['R_star_min'] = 0.13
        warnings.warn("tls_kwargs: R_star_min was undefined in TICv8. Filling it with R_star_min=0.13")
    if np.isnan(tls_kwargs['R_star_max']): 
        tls_kwargs['R_star_max'] = 3.5
        warnings.warn("tls_kwargs: R_star_max was undefined in TICv8. Filling it with R_star_max=3.5")
    if np.isnan(tls_kwargs['M_star']): 
        tls_kwargs['M_star'] = 1.
        warnings.warn("tls_kwargs: M_star was undefined in TICv8. Filling it with M_star=1.")
    if np.isnan(tls_kwargs['M_star_min']): 
        tls_kwargs['M_star_min'] = 0.1
        warnings.warn("tls_kwargs: M_star_min was undefined in TICv8. Filling it with M_star_min=0.1")
    if np.isnan(tls_kwargs['M_star_max']): 
        tls_kwargs['M_star_max'] = 1.
        warnings.warn("tls_kwargs: M_star_max was undefined in TICv8. Filling it with M_star_max=0.1")
    if np.isnan(tls_kwargs['u']).any(): 
        tls_kwargs['u'] = [0.4804, 0.1867]
        warnings.warn("tls_kwargs: u was undefined in TICv8. Filling it with u=[0.4804, 0.1867]")

    return tls_kwargs



###############################################################################
#::: write TLS reuslts as a dictionary to a json file
###############################################################################
def write_tls_results(fname, results):
    '''
    Parameters
    ----------
    fname : str
        Name of the output json file.
    results : transitleastsuqares.results class
        The results returned form a TLS run.

    Returns
    -------
    None.

    Outputs
    -------
    A json file that contains a dictionary of the most important tls results.
    The json file can be read into Python again via allesfitter's read_dic.
    
    Explanation
    -----------
    The TLS results object contains the following keys, where 
    'short' indicates it's a float or short list (e.g., the found period or depth per transit) and 
    'long' indicates it's a humongous array (e.g., the whole light curve).
    We only want to save the 'short' parts to save space:
        SDE short
        SDE_raw short
        chi2_min short
        chi2red_min short
        period short
        period_uncertainty short
        T0 short
        duration short
        depth short
        depth_mean short
        depth_mean_even short
        depth_mean_odd short
        transit_depths short
        transit_depths_uncertainties short
        rp_rs short
        snr short
        snr_per_transit short
        snr_pink_per_transit short
        odd_even_mismatch short
        transit_times short
        per_transit_count short
        transit_count short
        distinct_transit_count short
        empty_transit_count short
        FAP short
        in_transit_count short
        after_transit_count short
        before_transit_count short
        periods long
        power long
        power_raw long
        SR long
        chi2 long
        chi2red long
        model_lightcurve_time long
        model_lightcurve_model long
        model_folded_phase long
        folded_y long
        folded_dy long
        folded_phase long
        model_folded_model long
    Also:
        correct_duration short
        model long (our self-made model(time) curve)
    '''
    dic = {}
    for key in ['SDE', 'SDE_raw', 'chi2_min', 'chi2red_min', 'period', 'period_uncertainty',\
                'T0', 'duration', 'depth', 'depth_mean', 'depth_mean_even', 'depth_mean_odd',\
                'transit_depths', 'transit_depths_uncertainties', 'rp_rs',\
                'snr', 'snr_per_transit', 'snr_pink_per_transit', 'odd_even_mismatch',\
                'transit_times', 'per_transit_count', 'transit_count', 'distinct_transit_count',\
                'empty_transit_count', 'FAP', 'in_transit_count', 'after_transit_count',\
                'before_transit_count'] + ['correct_duration']: 
        if (type(results[key])!=np.ndarray): #if it's not an array, save it as is
            dic[key] = results[key]
        else:  #if it's a short array, save as list (for json)
            dic[key] = results[key].tolist()
    write_json(fname, dic)
    
    
    
###############################################################################
#::: function to convert the results into a dictionary
###############################################################################
def _to_dic(results):
    dic = {}
    for key in results: 
        dic[key] = results[key]
    return dic



###############################################################################
#::: TLS search on an input lightcurve
###############################################################################
def tls_search(time, flux, flux_err, plot=True, plot_type='brokenplot', **kwargs):
    '''
    Summary:
    -------
    This runs TLS on these data with the given infos
    
    Inputs:
    -------
    time : array of flaot
        time stamps of observations
    flux : array of flaot
        normalized flux
    flux_err : array of flaot
        error of normalized flux
    **kwargs : collection of keyword arguments
        All keyword arguments will be passed to TLS.
        Missing keywords will be replaced with default values:
            R_star : float
                radius of the star (e.g. median)
                default 1 R_sun (from TLS)
            R_star_min : float
                minimum radius of the star (e.g. 1st percentile)
                default 0.13 R_sun (from TLS)
            R_star_max : float
                maximum radius of the star (e.g. 99th percentile)
                default 3.5 R_sun (from TLS)
            M_star : float
                mass of the star (e.g. median)
                default 1. M_sun (from TLS)
            M_star_min : float
                minimum mass of the star (e.g. 1st percentile)
                default 0.1 M_sun (from TLS)
            M_star_max : float
                maximum mass of the star (e.g. 99th percentile)
                default 1. M_sun (from TLS)    
            u : list
                quadratic limb darkening parameters
                default [0.4804, 0.1867]
            period_min : float
                the minimum period to be searched (in days)
            period_max : float
                the maximum period to be searched (in days)
            show_progress_bar : bool
                Show a progress bar for TLS
                default True
            SNR_threshold : float
                the SNR threshold at which to stop the TLS search
                default 5
            SDE_threshold : float
                the SDE threshold at which to stop the TLS search
                default -inf
            FAP_threshold : float
                the False Alarm Probability threshold at which to stop the TLS search
                default inf
            quiet : bool
                silence all TLS outprint
                default True
        
    Returns:
    -------
    results_all : list of dictionaries
        List of all dictionaries containing the TLS results 
        (with dictionaries made from the transitleastsqaures.results class).
    fig_all : list of matplotlib.figure object, optional
        List of all summary figures. Only returned if plot is True.
    '''
    
    #::: seeed
    np.random.seed(42)
    
    
    #::: handle inputs
    time, flux, flux_err = clean(time, flux, flux_err)
    plot_bool = plot
        
    if 'show_progress_bar' not in kwargs: kwargs['show_progress_bar'] = True
    if 'SNR_threshold' not in kwargs: kwargs['SNR_threshold'] = 5.
    if 'SDE_threshold' not in kwargs: kwargs['SDE_threshold'] = -np.inf #don't trust SDE
    if 'FAP_threshold' not in kwargs: kwargs['FAP_threshold'] = np.inf #don't trust FAP 
    if 'quiet' not in kwargs: kwargs['quiet'] = True
    if 'inj_period' not in kwargs: kwargs['inj_period'] = np.nan
    
    non_tls_keys = ['SNR_threshold','SDE_threshold','FAP_threshold','quiet','inj_period']
    tls_kwargs_original = {key: kwargs[key] for key in kwargs.keys() if key not in non_tls_keys} #for the original tls
    #the rest is filled automatically by TLS if it was not given
    print('tls_kwargs_original', tls_kwargs_original)
    
    #::: init
    SNR = 1e12
    SDE = 1e12
    FAP = 0
    FOUND_SIGNAL = False
    results_all = []     
    fig_lightcurve_all = []     
    fig_folded_all = []        
    
    
    #::: function to run it once
    def _run1(time, flux, flux_err):
        if kwargs['quiet']:
            with open(os.devnull, 'w') as devnull:
                with contextlib.redirect_stdout(devnull):
                    with warnings.catch_warnings():
                        warnings.simplefilter("ignore")
                        model = tls(time, flux, flux_err)
                        results = model.power(**tls_kwargs_original)
        else:
            model = tls(time, flux, flux_err)
            results = model.power(**tls_kwargs_original)
        
        results = _to_dic(results)
        results['detection'] = (results['snr'] >= kwargs['SNR_threshold']) and (results['SDE'] >= kwargs['SDE_threshold']) and (results['FAP'] <= kwargs['FAP_threshold'])
        results['correct_duration'] = np.nan        
        results['R_planet_'] = np.nan

        
        if results['detection']:
            #::: calculcate the correct_duration, as TLS sometimes returns unreasonable durations
            ind_tr_phase = np.where( results['model_folded_model'] < 1. )[0]
            results['correct_duration'] = results['period'] * (results['model_folded_phase'][ind_tr_phase[-1]] - results['model_folded_phase'][ind_tr_phase[0]])
            
            if 'R_star' in kwargs:
                results['R_planet'] = results['rp_rs'] * kwargs['R_star'] * 109.07637070600963 #from Rsun to Rearth
            
        return results
            
            
    #::: function to plot it once
    # def _plot1(time, flux, flux_err, results):
    #     fig, axes = plt.subplots(1, 3, figsize=(20,5), tight_layout=True)
        
    #     ax = axes[0]
    #     ax.plot(results['folded_phase'], results['folded_y'], 'k.', color='silver', rasterized=True)
    #     bintime, binflux, binflux_err, _ = rebin_err(results['folded_phase'], results['folded_y'], dt = 0.001*results['period'], ferr_type='medsig', ferr_style='sem')
    #     ax.plot(bintime, binflux, 'b.', rasterized=True)
    #     ax.plot(results['model_folded_phase'], results['model_folded_model'], 'r-', lw=3)
        
    #     ax = axes[1]
    #     ax.plot((results['folded_phase']-0.5)*results['period']*24, results['folded_y'], 'k.', color='silver', rasterized=True)
    #     bintime, binflux, binflux_err, _ = rebin_err((results['folded_phase']-0.5)*results['period']*24, results['folded_y'], dt = 0.001*results['period']*24, ferr_type='medsig', ferr_style='sem')
    #     ax.plot(bintime, binflux, 'bo', rasterized=True)
    #     ax.plot((results['model_folded_phase']-0.5)*results['period']*24, results['model_folded_model'], 'r-', lw=3)
    #     ax.set(xlim=[ -1.5*results['correct_duration']*24, +1.5*results['correct_duration']*24 ], xlabel='Time (h)', yticks=[])
        
    #     ax = axes[2]
    #     ax.text( .02, 0.95, 'P = ' + np.format_float_positional(results['period'],4) + ' d', ha='left', va='center', transform=ax.transAxes )
    #     ax.text( .02, 0.85, 'Depth = ' + np.format_float_positional(1e3*(1.-results['depth']),4) + ' ppt', ha='left', va='center', transform=ax.transAxes )
    #     ax.text( .02, 0.75, 'Duration = ' + np.format_float_positional(24*results['correct_duration'],4) + ' h', ha='left', va='center', transform=ax.transAxes )
    #     ax.text( .02, 0.65, 'T_0 = ' + np.format_float_positional(results['T0'],4) + ' d', ha='left', va='center', transform=ax.transAxes )
    #     ax.text( .02, 0.55, 'SNR = ' + np.format_float_positional(results['snr'],4), ha='left', va='center', transform=ax.transAxes )
    #     ax.text( .02, 0.45, 'SDE = ' + np.format_float_positional(results['SDE'],4), ha='left', va='center', transform=ax.transAxes )
    #     ax.text( .02, 0.35, 'FAP = ' + np.format_float_scientific(results['FAP'],4), ha='left', va='center', transform=ax.transAxes )
    #     ax.set_axis_off()
        
    #     return fig
            
            
    #::: search for transits in a loop
    while (SNR >= kwargs['SNR_threshold']) and (SDE >= kwargs['SDE_threshold']) and (FAP <= kwargs['FAP_threshold']) and (FOUND_SIGNAL==False):
       
        #::: run once 
        results = _run1(time, flux, flux_err)
        
        #::: if a transit was detected, store the results, plot, and apply a mask for the next run
        if results['detection']:
            results_all.append(results)
            
            results['model'] = np.interp(time, results['model_lightcurve_time'], results['model_lightcurve_model'])
            
            if plot_bool:
                # fig = _plot1(time, flux, flux_err, results)
                fig_lightcurve = _tls_search_plot_lightcurve(time, flux, results, typ=plot_type)
                fig_folded = _tls_search_plot_folded(time, flux, results)
                fig_lightcurve_all.append(fig_lightcurve)
                fig_folded_all.append(fig_folded)
           
            time, flux, flux_err = mask(time, flux, flux_err, 
                                        results['period'], 
                                        np.max((1.5*results['correct_duration'])), 
                                        results['T0'])

        #::: update values
        SNR = results['snr']
        SDE = results['SDE']
        FAP = results['FAP']
        if is_multiple_of(results['period'],kwargs['inj_period']): SNR = -np.inf #if run as part of an injection-recovery test, then abort if it matches the injected period
        
                
    #::: return
    if plot_bool:
        return results_all, fig_lightcurve_all, fig_folded_all
    else:
        return results_all



def _cut(time, model_lightcurve_time, model_lightcurve_flux):
    return np.interp(time, model_lightcurve_time, model_lightcurve_flux) 



def _tls_search_plot_lightcurve(time, flux, results, typ='fullplot'):
    """
    ...

    Parameters
    ----------
    time : TYPE
        DESCRIPTION.
    flux : TYPE
        DESCRIPTION.
    flux_err : TYPE
        DESCRIPTION.
    results : TYPE
        DESCRIPTION.
    typ : TYPE, optional
        'fullplot', 'brokenplot', 'tessplot', 'monthplot'. The default is 'fullplot'.

    Returns
    -------
    None.
    """
    
    if typ=='fullplot':
        axes = fullplot(time, flux)
        axes = fullplot(results['model_lightcurve_time'], results['model_lightcurve_model'], color='r', ls='-', marker='', lw=3, zorder=100, axes=axes)
    elif typ=='brokenplot':
        axes = brokenplot(time, flux)
        axes = brokenplot(results['model_lightcurve_time'], results['model_lightcurve_model'], color='r', ls='-', marker='', lw=3, zorder=100, axes=axes)
    elif typ=='tessplot':
        trend = _cut(time, results['model_lightcurve_time'], results['model_lightcurve_model'])
        axes = tessplot(time, flux, trend=trend)
        # axes = tessplot(results['model_lightcurve_time'], results['model_lightcurve_model'], color='r', ls='-', marker='', lw=3, zorder=100, axes=axes, shade=False)
    elif typ=='monthplot':
        axes = monthplot(time, flux)
        axes = monthplot(results['model_lightcurve_time'], results['model_lightcurve_model'], color='r', ls='-', marker='', lw=3, zorder=100, axes=axes)
    
    return plt.gcf()
    

    
def _tls_search_plot_folded(time, flux, results):
    """
    ...

    Parameters
    ----------
    time : TYPE
        DESCRIPTION.
    flux : TYPE
        DESCRIPTION.
    results : TYPE
        DESCRIPTION.
    axes : TYPE, optional
        DESCRIPTION. The default is None.

    Returns
    -------
    axes : TYPE
        DESCRIPTION.
    """
    
    fig, axes = plt.subplots(1, 3, figsize=(12,3), tight_layout=True)
    
    ax = axes[0]
    bintime, binflux, binflux_err, _ = rebin_err(results['folded_phase'], results['folded_y'], dt = 0.001*results['period'], ferr_type='medsig', ferr_style='sem')
    ax.plot(bintime, binflux, 'b.', rasterized=True)
    ax.plot(results['model_folded_phase'], results['model_folded_model'], 'r-', lw=3)
    ylim = ax.get_ylim()
    ax.plot(results['folded_phase'], results['folded_y'], 'k.', color='silver', rasterized=True, zorder=-1)
    ax.set_ylim(ylim)
    
    ax = axes[1]
    bintime, binflux, binflux_err, _ = rebin_err((results['folded_phase']-0.5)*results['period']*24, results['folded_y'], dt = 0.001*results['period']*24, ferr_type='medsig', ferr_style='sem')
    ax.plot(bintime, binflux, 'bo', rasterized=True)
    ax.plot((results['model_folded_phase']-0.5)*results['period']*24, results['model_folded_model'], 'r-', lw=3)
    ax.set(xlim=[ -1.5*results['correct_duration']*24, +1.5*results['correct_duration']*24 ], xlabel='Time (h)', yticks=[])
    ylim = ax.get_ylim()
    ax.plot((results['folded_phase']-0.5)*results['period']*24, results['folded_y'], 'k.', color='silver', rasterized=True, zorder=-1)
    ax.set_ylim(ylim)
    
    ax = axes[2]
    ax.text( .02, 0.95, 'P = ' + np.format_float_positional(results['period'],4) + ' d', ha='left', va='center', transform=ax.transAxes )
    ax.text( .02, 0.85, 'Depth = ' + np.format_float_positional(1e3*(1.-results['depth']),4) + ' ppt', ha='left', va='center', transform=ax.transAxes )
    ax.text( .02, 0.75, 'Duration = ' + np.format_float_positional(24*results['correct_duration'],4) + ' h', ha='left', va='center', transform=ax.transAxes )
    ax.text( .02, 0.65, 'T_0 = ' + np.format_float_positional(results['T0'],4) + ' d', ha='left', va='center', transform=ax.transAxes )
    ax.text( .02, 0.55, 'SNR = ' + np.format_float_positional(results['snr'],4), ha='left', va='center', transform=ax.transAxes )
    ax.text( .02, 0.45, 'SDE = ' + np.format_float_positional(results['SDE'],4), ha='left', va='center', transform=ax.transAxes )
    ax.text( .02, 0.35, 'FAP = ' + np.format_float_scientific(results['FAP'],4), ha='left', va='center', transform=ax.transAxes )
    ax.text( .02, 0.25, 'R_planet/R_star = ' + np.format_float_positional(results['rp_rs'],4), ha='left', va='center', transform=ax.transAxes )
    if ~np.isnan(results['R_planet']): 
        ax.text( .02, 0.15, 'R_planet = ' + np.format_float_positional(results['R_planet'],4), ha='left', va='center', transform=ax.transAxes )
    ax.set_axis_off()
    
    return fig
    


def _tls_search_plot_individual(time, flux, flux_err, results):
    pass #TODO

    

###############################################################################
#::: Convenient wrapper for TESS tasks
###############################################################################
#TODO: work in progress
def tls_search_tess(time, flux, flux_err, 
                     wotan_kwargs=None,
                     tls_kwargs=None,
                     bad_regions=None,
                     options=None):

    if options is None: options = {}
    if 'outdir' not in options: options['outdir'] = ''
    if wotan_kwargs is None: wotan_kwargs = {'flatten': {'method':'biweight', 'window_length':1}}
    
    #::: logprint
    with open( os.path.join(options['outdir'], 'logfile.log'), 'w' ) as f:
        f.write('TLS search, UTC ' + datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S') + '\n')
    logprint('\nWotan kwargs:', options=options)
    logpprint(wotan_kwargs, options=options)
    logprint('\nTLS kwargs:', options=options)
    logpprint(tls_kwargs, options=options)
    logprint('\nOptions:', options=options)
    logpprint(options, options=options)
    
    
    #::: cleaning
    flux_clean, fig1, fig2, fig3 = \
        tessclean(time, flux, plot=True,
                  method=wotan_kwargs['flatten']['method'],
                  window_length=wotan_kwargs['flatten']['window_length'],
                  bad_regions=bad_regions)
    
    write_csv(os.path.join(options['outdir'],'flux_clean.csv'), (time, flux_clean, flux_err), header='time,flux_clean,flux_err')
    
    with PdfPages( os.path.join(options['outdir'],'flux_clean.pdf') ) as pdf:
        pdf.savefig( fig1 )
        pdf.savefig( fig2 )
        pdf.savefig( fig3 )
    plt.close('all')
    
    
    #::: transit search
    results_all, fig_lightcurve_all, fig_folded_all = \
        tls_search(time, flux_clean, flux_err, 
                   plot=True, plot_type='tessplot',
                   **tls_kwargs)
    
    if len(results_all)>0:
        with open( os.path.join(options['outdir'],'tls_summary.txt'), 'w' ) as f:
            f.write('TLS found '+str(len(results_all))+' potential signal(s).')
            
        for i, results in enumerate(results_all):
            write_tls_results( os.path.join(options['outdir'],'tls_signal_'+str(i)+'.txt'), results )
            
        for i, (fig1, fig2) in enumerate(zip(fig_lightcurve_all, fig_folded_all)):
            with PdfPages( os.path.join(options['outdir'],'tls_signal_'+str(i)+'.pdf') ) as pdf:
                pdf.savefig( fig1 )
                pdf.savefig( fig2 )
        plt.close('all')
        
    else:
        with open( os.path.join(options['outdir'],'tls_summary.txt'), 'w' ) as f:
            f.write('TLS found no potential signal(s).')
        
            

###############################################################################
#::: TLS search on an input lightcurve
###############################################################################
# def tls_search_old(time, flux, flux_err,
#                known_transits=None,
#                tls_kwargs=None,
#                wotan_kwargs=None,
#                options=None):
#     '''
#     Summary:
#     -------
#     This runs TLS on these data with the given infos
    
#     Inputs:
#     -------
#     time : array of flaot
#         time stamps of observations
#     flux : array of flaot
#         normalized flux
#     flux_err : array of flaot
#         error of normalized flux
        
#     Optional Inputs:
#     ----------------
#     known_transits : None or dict
#         >> can be used to mask known transits before running TLS
#         if None
#             nothing happens
#         if dict 
#             if one transit is already known, give for example: 
#                 known_transits = {'period':[1.3], 'duration':[2.1], 'epoch':[245800.0]}
#             if multiple transits are already known, give for example: 
#                 known_transits = {'name':['b','c'], 'period':[1.3, 21.0], 'duration':[2.1, 4.1], 'epoch':[245800.0, 245801.0]}
#             'period' is the period of the known transit(s)
#             'duration' is the total duration of the known transit(s), i.e. from first ingress point to last egrees point, in days
#             'epoch' is the epoch of the known transit(s)
        
#     tls_kwargs : None, str or dict:
#         >> can be used to fine-tune the TLS algorithm
#         if None
#             the default parameters will be chosen (see below)
#         if 'default'
#             the default parameters will be chosen (see below)
#         if dict
#             a dictionary with the following keywords is expected; 
#             missing keywords will be replaced with default values
#             R_star : float
#                 radius of the star (e.g. median)
#                 default 1 R_sun (from TLS)
#             R_star_min : float
#                 minimum radius of the star (e.g. 1st percentile)
#                 default 0.13 R_sun (from TLS)
#             R_star_max : float
#                 maximum radius of the star (e.g. 99th percentile)
#                 default 3.5 R_sun (from TLS)
#             M_star : float
#                 mass of the star (e.g. median)
#                 default 1. M_sun (from TLS)
#             M_star_min : float
#                 minimum mass of the star (e.g. 1st percentile)
#                 default 0.1 M_sun (from TLS)
#             M_star_max : float
#                 maximum mass of the star (e.g. 99th percentile)
#                 default 1. M_sun (from TLS)    
#             u : list
#                 quadratic limb darkening parameters
#                 default [0.4804, 0.1867]
#             SNR_threshold : float
#                 the SNR threshold at which to stop the TLS search
#                 default 5
#             SDE_threshold : float
#                 the SDE threshold at which to stop the TLS search
#                 default -inf
#             FAP_threshold : float
#                 the False Alarm Probability threshold at which to stop the TLS search
#                 default inf
#             period_min : float
#                 the minimum period to be searched (in days)
#             period_max : float
#                 the maximum period to be searched (in days)
        
#     wotan_kwargs : None, str, or dict:
#         >> can be used to detrend the data before the TLS search
#         if None
#             the default detrending will run (see below)
#         if str is 'default'
#             the default detrending will run (see below)
#         if str is 'off'
#             no detrending will run
#         if dict
#             a dictionary with two sub-dictionaries is expected; 
#             missing keywords will be replaced with default values
#             wotan_kwargs['slide_clip'] : dict
#                 this dictionary contains all slide clipping arguments
#                 window_length : float
#                     slide clip window length
#                     default 1
#                 low : float
#                     slide clip lower sigma
#                     default 20
#                 high : float
#                     slide clip upper sigma
#                     default 3
#             wotan_kwargs['flatten'] : dict
#                 this dictionary contains contains all detrending arguments
#                 method : str
#                     detrending method
#                     default 'biweight'
#                 window_length : float
#                     detrending window length in days
#                     default 1         
            
#     options : None or dict, keywords:
#         >> can be used for any general options
#         if None
#             the default options will be used (see below)
#         if dict
#             a dictionary with the following keywords is expected;
#             missing keywords will be replaced with default values
#             show_plot : bool
#                 can show a plot of each phase-folded transit candidate and TLS model in the terminal 
#                 default is True
#             save_plot : bool or str
#                 can save a plot of each phase-folded transit candidate and TLS model into outdir
#                 if True, will be set to '123'
#                 if str, then: '1': detrended plot, '2': TLS plot, '3': all TLS plots, and any combinations thereof
#                 default is True
#             save_csv : bool
#                 can save a csv of the detrended lightcurve
#                 default is True
#             outdir : string
#                 if None
#                     a new directory called "results" will be created in the current folder
#                 default is "tls_results_[wotan_flatten_method]_[wotan_flatten_window_length]"
        
#     Returns:
#     -------
#     List of all TLS results
#     '''
    
#     #::: seeed
#     np.random.seed(42)
    
    
#     #::: handle inputs
#     def clean(time,flux,flux_err):
#         if flux_err is None:
#             ind = np.where( ~np.isnan(time*flux) )[0]
#             time = time[ind]
#             flux = flux[ind]
#         else:
#             ind = np.where( ~np.isnan(time*flux*flux_err) )[0]
#             time = time[ind]
#             flux = flux[ind]
#             flux_err = flux_err[ind]
#         return time, flux, flux_err
    
#     time, flux, flux_err = clean(time,flux,flux_err)
#     time_input = 1.*time
#     flux_input = 1.*flux #for plotting
        
    
#     if type(wotan_kwargs)==str and wotan_kwargs=='off': 
#         detrend = False
#     else:
#         detrend = True
#         if (wotan_kwargs is None) or (type(wotan_kwargs)==str and wotan_kwargs=='default'): wotan_kwargs={} 
#         if 'slide_clip' not in wotan_kwargs: wotan_kwargs['slide_clip'] = {}
#         if wotan_kwargs['slide_clip'] is not None:
#             if 'window_length' not in wotan_kwargs['slide_clip']: wotan_kwargs['slide_clip']['window_length'] = 1.
#             if 'low' not in wotan_kwargs['slide_clip']: wotan_kwargs['slide_clip']['low'] = 20.
#             if 'high' not in wotan_kwargs['slide_clip']: wotan_kwargs['slide_clip']['high'] = 3.
    
#     if 'flatten' not in wotan_kwargs: wotan_kwargs['flatten'] = {}
#     if wotan_kwargs['flatten'] is not None:
#         if 'method' not in wotan_kwargs['flatten']: wotan_kwargs['flatten']['method'] = 'biweight'
#         if 'window_length' not in wotan_kwargs['flatten']: wotan_kwargs['flatten']['window_length'] = 1.
#     #the rest is filled automatically by Wotan
        
#     if tls_kwargs is None: tls_kwargs = {}
#     if 'show_progress_bar' not in tls_kwargs: tls_kwargs['show_progress_bar'] = False
#     if 'SNR_threshold' not in tls_kwargs: tls_kwargs['SNR_threshold'] = 5.
#     if 'SDE_threshold' not in tls_kwargs: tls_kwargs['SDE_threshold'] = -np.inf #don't trust SDE
#     if 'FAP_threshold' not in tls_kwargs: tls_kwargs['FAP_threshold'] = np.inf #don't trust FAP 
#     tls_kwargs_original = {key: tls_kwargs[key] for key in tls_kwargs.keys() if key not in ['SNR_threshold','SDE_threshold','FAP_threshold']} #for the original tls
#     #the rest is filled automatically by TLS
    
#     if options is None: options = {}
#     if 'show_plot' not in options: options['show_plot'] = True
#     if type(options['show_plot'])==bool and (options['show_plot'] is True): options['show_plot']='123' #1: detrended plot, 2: TLS plot, 3: all TLS plots
#     if type(options['show_plot'])==bool and (options['show_plot'] is False): options['show_plot']='' #1: detrended plot, 2: TLS plot, 3: all TLS plots
#     if 'save_plot' not in options: options['save_plot'] = True
#     if type(options['save_plot'])==bool and (options['save_plot'] is True): options['save_plot']='123' #1: detrended plot, 2: TLS plot, 3: all TLS plots
#     if type(options['save_plot'])==bool and (options['save_plot'] is False): options['save_plot']='' #1: detrended plot, 2: TLS plot, 3: all TLS plots
#     if 'save_csv' not in options: options['save_csv'] = True
#     if 'outdir' not in options: 
#         if detrend:
#             options['outdir'] = 'tls_results_'+wotan_kwargs['flatten']['method']+'_'+str(wotan_kwargs['flatten']['window_length'])
#         else:
#             options['outdir'] = 'tls_results_undetrended'
#     if 'quiet' not in options: options['quiet'] = True
#     if 'inj_period' not in options: options['inj_period'] = np.nan
    
    
#     #::: init
#     SNR = 1e12
#     SDE = 1e12
#     FAP = 0
#     FOUND_SIGNAL = False
#     results_all = []      
#     if len(options['outdir'])>0 and not os.path.exists(options['outdir']): os.makedirs(options['outdir'])  
    
    
#     #::: logprint
#     with open( os.path.join(options['outdir'], 'logfile.log'), 'w' ) as f:
#         f.write('TLS search, UTC ' + datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S') + '\n')
#     logprint('\nWotan kwargs:', options=options)
#     logpprint(wotan_kwargs, options=options)
#     logprint('\nTLS kwargs:', options=options)
#     logpprint(tls_kwargs, options=options)
#     logprint('\nOptions:', options=options)
#     logpprint(options, options=options)
    
#     # timer1 = timer.time()
#     # print('t1', timer1 - timer0)
            
#     #::: apply a mask (if wished so)
#     if known_transits is not None:
#         for period, duration, T0 in zip(known_transits['period'], known_transits['duration'], known_transits['epoch']):
#             time, flux, flux_err = mask(time, flux, flux_err, period, duration, T0)
    
    
#     #::: global sigma clipping
#     flux = sigma_clip(flux, sigma_upper=3, sigma_lower=20)
    
#     # timer2 = timer.time()
#     # print('t2', timer2 - timer0)
    
#     #::: detrend (if wished so)
#     if detrend:
        
#         #::: slide clipping (super slow)
#         # if wotan_kwargs['slide_clip'] is not None: flux = slide_clip(time, flux, **wotan_kwargs['slide_clip']) #slide_clip is super slow (10 seconds for a TESS 2 min lightcurve for a single Sector)
#         # timer3a = timer.time()
#         # print('t3a', timer3a - timer0)   
    
#         #::: fast slide clipping (super fast)
#         if wotan_kwargs['slide_clip'] is not None: 
#             flux = slide_clip(time, flux, **wotan_kwargs['slide_clip']) #slide_clip is super fast (<1 seconds for a TESS 2 min lightcurve for a single Sector)
#             flux_clip = 1*flux
#         # timer3a = timer.time()
#         # print('t3a', timer3a - timer0)   
        
#         #::: detrending (super fast)
#         if wotan_kwargs['flatten'] is not None: 
#             flux, trend = flatten(time, flux, return_trend=True, **wotan_kwargs['flatten']) #flatten is super fast, (<1 second for a TESS 2 min lightcurve for a single Sector)
#         # timer3b = timer.time()
#         # print('t3b', timer3b - timer0)   
        
#         #::: global sigma clipping on the flattened flux (super fast)
#         flux = sigma_clip(flux, sigma_upper=3, sigma_lower=20)
#         # timer3c = timer.time()
#         # print('t3c', timer3c - timer0)   
        
#         if ('1' in options['show_plot']) or ('1' in options['save_plot']):
#             gone = np.isnan(time_input*flux_input)
#             print(time_input, gone)
#             axes = tessplot(time_input[gone], flux_input[gone], color='r')
#             tessplot(time, flux_clip, trend=trend, axes=axes, shade=False)
#             for ax in axes: ax.set_ylabel('Flux\n(original)')
#             fig1 = plt.gcf()
            
#             axes = tessplot(time, flux_clip, trend=trend)
#             for ax in axes: ax.set_ylabel('Flux\n(clipped)')
#             fig2 = plt.gcf()
            
#             axes = tessplot(time, flux)
#             fig3 = plt.gcf()
#             for ax in axes: ax.set_ylabel('Flux\n(clipped & detrended)')
            
#             # fig, axes = plt.subplots(2,1, figsize=(40,8))
#             # brokenplot(time_input, flux_input, trend=trend, ax=axes[0])
#             # axes[0].set(ylabel='Flux (input)', xticklabels=[])
#             # brokenplot(time, trend, fmt='r-', ax=axes[0])
#             # axes[0].plot(time_input, flux_input, 'b.', rasterized=True)
#             # axes[0].plot(time, trend, 'r-', lw=2)
#             # brokenplot(time_input, flux_input, ax=axes[1], clip=True)
#             # brokenplot(time, trend, fmt='r-', ax=axes[1], clip=True)
#             # axes[1].set(ylabel='Flux (clipped)', xticklabels=[])
#             # brokenplot(time, flux, ax=axes[1])
#             # axes[1].plot(time, flux, 'b.', rasterized=True)
#             # axes[1].set(ylabel='Flux (detrended)', xlabel='Time (BJD)')
#             # axes[2].set(ylabel='Flux (detrended)')
#         if ('1' in options['save_plot']):
#             # try: 
#             f = os.path.join(options['outdir'],'flux_'+wotan_kwargs['flatten']['method']+'.pdf')
#             with PdfPages(f) as pdf:
#                 pdf.savefig( fig1 )
#                 pdf.savefig( fig2 )
#                 pdf.savefig( fig3 )
#             #     fig.savefig(os.path.join(options['outdir'],'flux_'+wotan_kwargs['flatten']['method']+'.pdf'), bbox_inches='tight') #some matplotlib versions crash when saving pdf...
#             # except: 
#             #     fig.savefig(os.path.join(options['outdir'],'flux_'+wotan_kwargs['flatten']['method']+'.jpg'), bbox_inches='tight') #some matplotlib versions need pillow for jpg (conda install pillow)...
        
#         if ('1' in options['show_plot']):
#             plt.show()
#         else:
#             plt.close('all')
                
#         if options['save_csv']:
#             if flux_err is None: flux_err0 = np.nan*time
#             else: flux_err0 = flux_err
#             X = np.column_stack((time, flux, flux_err0, trend))
#             np.savetxt(os.path.join(options['outdir'],'flux_'+wotan_kwargs['flatten']['method']+'.csv'), X, delimiter=',', header='time,flux_detrended,flux_err,trend')
        
#         time_detrended = 1.*time #just for plotting
#         flux_detrended = 1.*flux #just for plotting
        
#     # timer3d = timer.time()
#     # print('t3d', timer3d - timer0)    
    
    
#     #::: search for transits
#     i = 0
#     ind_trs = []
#     while (SNR >= tls_kwargs['SNR_threshold']) and (SDE >= tls_kwargs['SDE_threshold']) and (FAP <= tls_kwargs['FAP_threshold']) and (FOUND_SIGNAL==False):
        
#         if options['quiet']:
#             with open(os.devnull, 'w') as devnull:
#                 with contextlib.redirect_stdout(devnull):
#                     with warnings.catch_warnings():
#                         warnings.simplefilter("ignore")
#                         model = tls(time, flux, flux_err)
#                         results = model.power(**tls_kwargs_original)
#         else:
#             model = tls(time, flux, flux_err)
#             results = model.power(**tls_kwargs_original)
            
#         # timer4 = timer.time()
#         # print('t4', timer4 - timer0)  
        
#         # plt.figure()
#         # plt.plot(time, flux, 'b.')
#         # pprint(tls_kwargs_original)
#         # pprint(results)
#         # err
        
#         if (results['snr'] >= tls_kwargs['SNR_threshold']) and (results['SDE'] >= tls_kwargs['SDE_threshold']) and (results['FAP'] <= tls_kwargs['FAP_threshold']):
            
#             #::: calculcate the correct_duration, as TLS sometimes returns unreasonable durations
#             ind_tr_phase = np.where( results['model_folded_model'] < 1. )[0]
#             correct_duration = results['period'] * (results['model_folded_phase'][ind_tr_phase[-1]] - results['model_folded_phase'][ind_tr_phase[0]])
            
#             #::: mark transit
#             ind_tr, ind_out = index_transits(time_input, results['T0'], results['period'], correct_duration)
#             ind_trs.append(ind_tr)
            
#             #::: mask out detected transits and append results
#             time1, flux1 = 1*time, 1*flux #for plotting
#             time, flux, flux_err = mask(time, flux, flux_err, results['period'], np.max((1.5*correct_duration)), results['T0'])
#             results = _to_dic(results)
#             results['correct_duration'] = correct_duration
#             results_all.append(results)
            
#             #::: write TLS stats to file
#             write_tls_results(os.path.join(options['outdir'],'tls_signal_'+str(i)+'.txt'), results)
#             # with open(os.path.join(options['outdir'],'tls_signal_'+str(i)+'.txt'), 'wt') as out:
#             #     pprint(results, stream=out)
    
#             # timer5 = timer.time()
#             # print('t5', timer5 - timer0)  
    
#             #::: individual TLS plots
#             if ('2' in options['show_plot']) or ('2' in options['save_plot']):
#                 fig = plt.figure(figsize=(20,8), tight_layout=True)
#                 gs = fig.add_gridspec(2,3)
                
#                 ax = fig.add_subplot(gs[0,:])
#                 ax.plot(time1, flux1, 'k.', color='silver', rasterized=True)
#                 bintime, binflux, binflux_err, _ = rebin_err(time1, flux1, dt = 10./60/24, ferr_type='medsig', ferr_style='sem') #in 10 min intervals
#                 ax.plot(bintime, binflux, 'b.', rasterized=True)
#                 ax.plot(results['model_lightcurve_time'], results['model_lightcurve_model'], 'r-', lw=3)
#                 ax.set(xlabel='Time (BJD)', ylabel='Flux')
                
#                 ax = fig.add_subplot(gs[1,0])
#                 ax.plot(results['folded_phase'], results['folded_y'], 'k.', color='silver', rasterized=True)
#                 bintime, binflux, binflux_err, _ = rebin_err(results['folded_phase'], results['folded_y'], dt = 0.001*results['period'], ferr_type='medsig', ferr_style='sem')
#                 ax.plot(bintime, binflux, 'b.', rasterized=True)
#                 # plot_phase_folded_lightcurve(time1, flux1, results['period'], results['T0'], dt=0.002, ax=ax)
#                 ax.plot(results['model_folded_phase'], results['model_folded_model'], 'r-', lw=3)
#                 # ax.set(xlabel='Phase', ylabel='Flux')
                
#                 ax = fig.add_subplot(gs[1,1])
#                 ax.plot((results['folded_phase']-0.5)*results['period']*24, results['folded_y'], 'k.', color='silver', rasterized=True)
#                 bintime, binflux, binflux_err, _ = rebin_err((results['folded_phase']-0.5)*results['period']*24, results['folded_y'], dt = 0.001*results['period']*24, ferr_type='medsig', ferr_style='sem')
#                 ax.plot(bintime, binflux, 'bo', rasterized=True)
#                 # plot_phase_folded_lightcurve(time1*24, flux1, results['period']*24, results['T0'], ax=ax, dt=0.002)
#                 ax.plot((results['model_folded_phase']-0.5)*results['period']*24, results['model_folded_model'], 'r-', lw=3)
#                 ax.set(xlim=[ -1.5*correct_duration*24, +1.5*correct_duration*24 ], xlabel='Time (h)', yticks=[])
                
#                 ax = fig.add_subplot(gs[1,2])
#                 ax.text( .02, 0.95, 'P = ' + np.format_float_positional(results['period'],4) + ' d', ha='left', va='center', transform=ax.transAxes )
#                 ax.text( .02, 0.85, 'Depth = ' + np.format_float_positional(1e3*(1.-results['depth']),4) + ' ppt', ha='left', va='center', transform=ax.transAxes )
#                 ax.text( .02, 0.75, 'Duration = ' + np.format_float_positional(24*correct_duration,4) + ' h', ha='left', va='center', transform=ax.transAxes )
#                 ax.text( .02, 0.65, 'T_0 = ' + np.format_float_positional(results['T0'],4) + ' d', ha='left', va='center', transform=ax.transAxes )
#                 ax.text( .02, 0.55, 'SNR = ' + np.format_float_positional(results['snr'],4), ha='left', va='center', transform=ax.transAxes )
#                 ax.text( .02, 0.45, 'SDE = ' + np.format_float_positional(results['SDE'],4), ha='left', va='center', transform=ax.transAxes )
#                 ax.text( .02, 0.35, 'FAP = ' + np.format_float_scientific(results['FAP'],4), ha='left', va='center', transform=ax.transAxes )
#                 ax.set_axis_off()
#                 if ('2' in options['save_plot']):
#                     try: fig.savefig(os.path.join(options['outdir'],'tls_signal_'+str(i)+'.pdf'), bbox_inches='tight') #some matplotlib versions crash when saving pdf...
#                     except: fig.savefig(os.path.join(options['outdir'],'tls_signal_'+str(i)+'.jpg'), bbox_inches='tight') #some matplotlib versions need pillow for jpg (conda install pillow)...
#                 if ('2' in options['show_plot']):
#                     plt.show(fig)
#                 else:
#                     plt.close(fig)
                    
#             # timer6 = timer.time()
#             # print('t6', timer6 - timer0)  
            
#         SNR = results['snr']
#         SDE = results['SDE']
#         FAP = results['FAP']
#         if is_multiple_of(results['period'],options['inj_period']): SNR = -np.inf #if run as part of an inejction-recovery test, then abort if it matches the injected period
#         i+=1
        
        
        
#     #::: full lightcurve plot
#     if ('3' in options['show_plot']) or ('3' in options['save_plot']):
        
#         if detrend:
#             fig, axes = plt.subplots(2,1, figsize=(40,8), tight_layout=True)
#             ax = axes[0]
#             ax.plot(time_input, flux_input, 'k.', color='grey', rasterized=True)
#             ax.plot(time_input, trend, 'r-', lw=2)
#             for number, ind_tr in enumerate(ind_trs):
#                 ax.plot(time_input[ind_tr], flux_input[ind_tr], marker='.', linestyle='none', label='signal '+str(number))
#             ax.set(ylabel='Flux (input)', xticklabels=[])
#             ax.legend()

#             ax = axes[1]
#             ax.plot(time_detrended, flux_detrended, 'k.', color='grey', rasterized=True)
#             for number, ind_tr in enumerate(ind_trs):
#                 ax.plot(time_detrended[ind_tr], flux_detrended[ind_tr], marker='.', linestyle='none', label='signal '+str(number))
#             ax.set(ylabel='Flux (detrended)', xlabel='Time (BJD)')
#             ax.legend()
            
#         else:
#             fig = plt.figure(figsize=(20,4), tight_layout=True)
#             fig, ax = plt.subplots(1,1, figsize=(40,4))
#             ax.plot(time_input, flux_input, 'k.', color='grey', rasterized=True)
#             ax.set(ylabel='Flux (input)', xlabel='Time (BJD)')
#             for number, ind_tr in enumerate(ind_trs):
#                 ax.plot(time_input[ind_tr], flux_input[ind_tr], marker='.', linestyle='none', label='signal '+str(number))
#             ax.legend()
        
#         if ('3' in options['save_plot']):
#             try: fig.savefig(os.path.join(options['outdir'],'tls_signal_all.pdf'), bbox_inches='tight') #some matplotlib versions crash when saving pdf...
#             except: fig.savefig(os.path.join(options['outdir'],'tls_signal_all.jpg'), bbox_inches='tight') #some matplotlib versions need pillow for jpg (conda install pillow)...
#         if ('3' in options['show_plot']):
#             plt.show(fig)
#         else:
#             plt.close(fig)                    
                
            
#     return results_all



###############################################################################
#::: TLS search using tessio
###############################################################################
def tls_search_by_tic(tic_id,
                      tls_kwargs=None, SNR_threshold=5., known_transits=None,
                      options=None):
    '''
    Summary:
    -------
    wrapper around tls_search()
    retrieves the SPOC PDC-SAP lightcurve
    retrieves all TIC catalog information from MAST
    calls tls_search()
    
    Inputs:
    -------
    tic_id : str
        TIC ID
        
    Optional Inputs:
    ----------------
    see tls_search()
        
    Returns:
    -------
    list of all TLS results
    '''
    
    #::: handle inputs
    if options is None: options = {}
    if 'show_plot' not in options: options['show_plot']=False
    if 'save_plot' not in options: options['save_plot']=False
    if 'outdir' not in options: options['outdir']=''
    
    #::: format inputs
    tic_id = str(int(tic_id))
    
    #::: load data and inject transit
    time, flux, flux_err = tessio.get(tic_id, pipeline='spoc', PDC=True, unpack=True)
    
    #::: load TIC info / tls kwargs
    tls_kwargs = get_tls_kwargs_by_tic(tic_id, tls_kwargs=tls_kwargs)
    
    return tls_search(time, flux, flux_err,
                      tls_kwargs=tls_kwargs,
                      SNR_threshold=SNR_threshold,
                      known_transits=known_transits,
                      options=options)



###############################################################################
#::: main
###############################################################################
if __name__ == '__main__':
    pass
    
    ###########################################################################
    #::: Example: search for a transit with TLS and tessio
    ###########################################################################
    # tic_id = '269701147'
    # SNR_threshold=5.,
    # known_transits = {'epoch':[2458715.3547, 2458726.0526, 2458743.5534],
    #                   'period':[8.8806, 28.5810, 38.3497],
    #                   'duration':[3.09/24., 4.45/24., 5.52/24.]
    #                  }
    
    # results_all = tls_search_by_tic(tic_id,
    #                 SNR_threshold=SNR_threshold,
    #                 known_transits=known_transits)
    # print(results_all)